#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import abc

from toscaparser.common.exception import ExceptionCollector
from toscaparser.common.exception import UnknownInputError
from toscaparser.dataentity import DataEntity
from toscaparser.elements.constraints import Schema
from toscaparser.elements.datatype import DataType
from toscaparser.elements.entity_type import EntityType
from toscaparser.elements.relationshiptype import RelationshipType
from toscaparser.elements.statefulentitytype import StatefulEntityType
from toscaparser.utils.gettextutils import _


GET_PROPERTY = 'get_property'
GET_ATTRIBUTE = 'get_attribute'
GET_INPUT = 'get_input'
GET_OPERATION_OUTPUT = 'get_operation_output'
CONCAT = 'concat'
TOKEN = 'token'

SELF = 'SELF'
HOST = 'HOST'
TARGET = 'TARGET'
SOURCE = 'SOURCE'

HOSTED_ON = 'tosca.relationships.HostedOn'


class Function(object, metaclass=abc.ABCMeta):
    """An abstract type for representing a Tosca template function."""

    def __init__(self, tosca_tpl, context, name, args):
        self.tosca_tpl = tosca_tpl
        self.context = context
        self.name = name
        self.args = args
        self.validate()

    @abc.abstractmethod
    def result(self):
        """Invokes the function and returns its result

        Some methods invocation may only be relevant on runtime (for example,
        getting runtime properties) and therefore its the responsibility of
        the orchestrator/translator to take care of such functions invocation.

        :return: Function invocation result.
        """
        return {self.name: self.args}

    @abc.abstractmethod
    def validate(self):
        """Validates function arguments."""
        pass


class GetInput(Function):
    """Get a property value declared within the input of the service template.

    Arguments:

    * Input name.

    Example:

    * get_input: port
    """

    def validate(self):
        if len(self.args) != 1:
            ExceptionCollector.appendException(
                ValueError(_(
                    'Expected one argument for function "get_input" but '
                    'received "%s".') % self.args))
        inputs = [input.name for input in self.tosca_tpl.inputs]
        if self.args[0] not in inputs:
            ExceptionCollector.appendException(
                UnknownInputError(input_name=self.args[0]))

    def result(self):
        if self.tosca_tpl.parsed_params and \
           self.input_name in self.tosca_tpl.parsed_params:
            return DataEntity.validate_datatype(
                self.tosca_tpl.tpl['inputs'][self.input_name]['type'],
                self.tosca_tpl.parsed_params[self.input_name])

        input = [input_def for input_def in self.tosca_tpl.inputs
                 if self.input_name == input_def.name]
        if input:
            return input[0].default
        return None

    @property
    def input_name(self):
        return self.args[0]


class GetAttribute(Function):
    """Get an attribute value of an entity defined in the service template

    Node template attributes values are set in runtime and therefore its the
    responsibility of the Tosca engine to implement the evaluation of
    get_attribute functions.

    Arguments:

    * Node template name | HOST.
    * Attribute name.

    If the HOST keyword is passed as the node template name argument the
    function will search each node template along the HostedOn relationship
    chain until a node which contains the attribute is found.

    Examples:

    * { get_attribute: [ server, private_address ] }
    * { get_attribute: [ HOST, private_address ] }
    * { get_attribute: [ HOST, private_address, 0 ] }
    * { get_attribute: [ HOST, private_address, 0, some_prop] }
    """

    def validate(self):
        if len(self.args) < 2:
            ExceptionCollector.appendException(
                ValueError(_('Illegal arguments for function "{0}". Expected '
                             'arguments: "node-template-name", "req-or-cap"'
                             '(optional), "property name"'
                             ).format(GET_ATTRIBUTE)))
            return
        elif len(self.args) == 2:
            self._find_node_template_containing_attribute()
        else:
            node_tpl = self._find_node_template(self.args[0])
            if node_tpl is None:
                return
            index = 2
            attrs = node_tpl.type_definition.get_attributes_def()
            found = [attrs[self.args[1]]] if self.args[1] in attrs else []
            if found:
                attr = found[0]
            else:
                index = 3
                # then check the req or caps
                attr = self._find_req_or_cap_attribute(self.args[1],
                                                       self.args[2])
                if not attr:
                    return

            value_type = attr.schema['type']
            value_schema = attr.schema
            if len(self.args) > index:
                for elem in self.args[index:]:
                    if value_type == "list":
                        if not isinstance(elem, int):
                            ExceptionCollector.appendException(
                                ValueError(_('Illegal arguments for function'
                                             ' "{0}". "{1}" Expected positive'
                                             ' integer argument'
                                             ).format(GET_ATTRIBUTE, elem)))
                        value_type = value_schema['entry_schema']['type']
                    elif value_type == "map":
                        value_type = value_schema['entry_schema']['type']
                    elif value_type in Schema.PROPERTY_TYPES:
                        ExceptionCollector.appendException(
                            ValueError(_('Illegal arguments for function'
                                         ' "{0}". Unexpected attribute/'
                                         'index value "{1}"'
                                         ).format(GET_ATTRIBUTE, elem)))
                        return
                    else:  # It is a complex type
                        data_type = DataType(value_type,
                                             self.tosca_tpl.custom_defs)
                        props = data_type.get_all_properties()
                        found = [props[elem]] if elem in props else []
                        if found:
                            prop = found[0]
                            value_type = prop.schema['type']
                            value_schema = prop.schema
                        else:
                            ExceptionCollector.appendException(
                                KeyError(_('Illegal arguments for function'
                                           ' "{0}". Attribute name "{1}" not'
                                           ' found in "{2}"'
                                           ).format(GET_ATTRIBUTE,
                                                    elem,
                                                    value_type)))

    def result(self):
        return self

    def get_referenced_node_template(self):
        """Gets the NodeTemplate instance the get_attribute function refers to.

        If HOST keyword was used as the node template argument, the node
        template which contains the attribute along the HostedOn relationship
        chain will be returned.
        """
        return self._find_node_template_containing_attribute()

    # Attributes can be explicitly created as part of the type definition
    # or a property name can be implicitly used as an attribute name
    def _find_node_template_containing_attribute(self):
        node_tpl = self._find_node_template(self.args[0])
        if node_tpl and \
                not self._attribute_exists_in_type(node_tpl.type_definition) \
                and self.attribute_name not in node_tpl.get_properties():
            ExceptionCollector.appendException(
                KeyError(_('Attribute "%(att)s" was not found in node '
                           'template "%(ntpl)s".') %
                         {'att': self.attribute_name,
                          'ntpl': node_tpl.name}))
        return node_tpl

    def _attribute_exists_in_type(self, type_definition):
        attrs_def = type_definition.get_attributes_def()
        found = [attrs_def[self.attribute_name]] \
            if self.attribute_name in attrs_def else []
        return len(found) == 1

    def _find_host_containing_attribute(self, node_template_name=SELF):
        node_template = self._find_node_template(node_template_name)
        if node_template:
            hosted_on_rel = EntityType.TOSCA_DEF[HOSTED_ON]
            for r in node_template.requirements:
                for requirement, target_name in r.items():
                    target_node = self._find_node_template(target_name)
                    target_type = target_node.type_definition
                    for capability in target_type.get_capabilities_objects():
                        if capability.inherits_from(
                                hosted_on_rel['valid_target_types']):
                            if self._attribute_exists_in_type(target_type):
                                return target_node
                            return self._find_host_containing_attribute(
                                target_name)

    def _find_node_template(self, node_template_name):
        # if the node_template_name has the long format
        if isinstance(node_template_name, dict):
            # get only the node name
            if 'node' in node_template_name:
                node_template_name = node_template_name['node']
            else:
                ExceptionCollector.appendException(
                    ValueError(_(' No node name in the relationship.')))
                return
        if node_template_name == HOST:
            # Currently this is the only way to tell whether the function
            # is used within the outputs section of the TOSCA template.
            if isinstance(self.context, list):
                ExceptionCollector.appendException(
                    ValueError(_(
                        '"get_attribute: [ HOST, ... ]" is not allowed in '
                        '"outputs" section of the TOSCA template.')))
                return
            node_tpl = self._find_host_containing_attribute()
            if not node_tpl:
                ExceptionCollector.appendException(
                    ValueError(_(
                        '"get_attribute: [ HOST, ... ]" was used in node '
                        'template "{0}" but "{1}" was not found in '
                        'the relationship chain.').format(self.context.name,
                                                          HOSTED_ON)))
                return
            return node_tpl
        if node_template_name == TARGET:
            if not isinstance(self.context.type_definition, RelationshipType):
                ExceptionCollector.appendException(
                    KeyError(_('"TARGET" keyword can only be used in context'
                               ' to "Relationships" target node')))
                return
            return self.context.target
        if node_template_name == SOURCE:
            if not isinstance(self.context.type_definition, RelationshipType):
                ExceptionCollector.appendException(
                    KeyError(_('"SOURCE" keyword can only be used in context'
                               ' to "Relationships" source node')))
                return
            return self.context.source
        name = self.context.name \
            if node_template_name == SELF and \
            not isinstance(self.context, list) \
            else node_template_name
        for node_template in self.tosca_tpl.nodetemplates:
            if node_template.name == name:
                return node_template
        ExceptionCollector.appendException(
            KeyError(_(
                'Node template "{0}" was not found.'
            ).format(node_template_name)))

    def _find_req_or_cap_attribute(self, req_or_cap, attr_name):
        node_tpl = self._find_node_template(self.args[0])
        # Find attribute in node template's requirements
        for r in node_tpl.requirements:
            for req, node_name in r.items():
                if req == req_or_cap:
                    node_template = self._find_node_template(node_name)
                    return self._get_capability_attribute(
                        node_template,
                        req,
                        attr_name)
        # If requirement was not found, look in node template's capabilities
        return self._get_capability_attribute(node_tpl,
                                              req_or_cap,
                                              attr_name)

    def _get_capability_attribute(self,
                                  node_template,
                                  capability_name,
                                  attr_name):
        """Gets a node template capability attribute."""
        caps = node_template.get_capabilities()
        if caps and capability_name in caps.keys():
            cap = caps[capability_name]
            attribute = None
            attrs = cap.definition.get_attributes_def()
            if attrs and attr_name in attrs.keys():
                attribute = attrs[attr_name]
            if not attribute:
                ExceptionCollector.appendException(
                    KeyError(_('Attribute "%(attr)s" was not found in '
                               'capability "%(cap)s" of node template '
                               '"%(ntpl1)s" referenced from node template '
                               '"%(ntpl2)s".') % {'attr': attr_name,
                                                  'cap': capability_name,
                                                  'ntpl1': node_template.name,
                                                  'ntpl2': self.context.name}))
            return attribute
        msg = _('Requirement/Capability "{0}" referenced from node template '
                '"{1}" was not found in node template "{2}".').format(
                    capability_name,
                    self.context.name,
                    node_template.name)
        ExceptionCollector.appendException(KeyError(msg))

    @property
    def node_template_name(self):
        return self.args[0]

    @property
    def attribute_name(self):
        return self.args[1]


class GetProperty(Function):
    """Get a property value of an entity defined in the same service template.

    Arguments:

    * Node template name | SELF | HOST | SOURCE | TARGET.
    * Requirement or capability name (optional).
    * Property name.

    If requirement or capability name is specified, the behavior is as follows:
    The req or cap name is first looked up in the specified node template's
    requirements.
    If found, it would search for a matching capability
    of an other node template and get its property as specified in function
    arguments.
    Otherwise, the req or cap name would be looked up in the specified
    node template's capabilities and if found, it would return  the property of
    the capability as specified in function arguments.

    Examples:

    * { get_property: [ mysql_server, port ] }
    * { get_property: [ SELF, db_port ] }
    * { get_property: [ SELF, database_endpoint, port ] }
    * { get_property: [ SELF, database_endpoint, port, 1 ] }
    """

    def validate(self):
        if len(self.args) < 2:
            ExceptionCollector.appendException(
                ValueError(_(
                    'Expected arguments: "node-template-name", "req-or-cap" '
                    '(optional), "property name".')))
            return
        if len(self.args) == 2:
            found_prop = self._find_property(self.args[1])
            if not found_prop:
                return
            prop = found_prop.value
            if not isinstance(prop, Function):
                get_function(self.tosca_tpl, self.context, prop)
        elif len(self.args) >= 3:
            # do not use _find_property to avoid raise KeyError
            # if the prop is not found
            # First check if there is property with this name
            node_tpl = self._find_node_template(self.args[0])
            props = node_tpl.get_properties() if node_tpl else []
            index = 2
            found = [props[self.args[1]]] if self.args[1] in props else []
            if found:
                property_value = found[0].value
            else:
                index = 3
                # then check the req or caps
                property_value = self._find_req_or_cap_property(self.args[1],
                                                                self.args[2])
            if len(self.args) > index:
                for elem in self.args[index:]:
                    if isinstance(property_value, list):
                        int_elem = int(elem)
                        property_value = self._get_index_value(property_value,
                                                               int_elem)
                    else:
                        property_value = self._get_attribute_value(
                            property_value,
                            elem)

    def _find_req_or_cap_property(self, req_or_cap, property_name):
        node_tpl = self._find_node_template(self.args[0])
        if node_tpl is None:
            return None
        # Find property in node template's requirements
        for r in node_tpl.requirements:
            for req, node_name in r.items():
                if req == req_or_cap:
                    node_template = self._find_node_template(node_name)
                    return self._get_capability_property(
                        node_template,
                        req,
                        property_name)
        # If requirement was not found, look in node template's capabilities
        return self._get_capability_property(node_tpl,
                                             req_or_cap,
                                             property_name)

    def _get_capability_property(self,
                                 node_template,
                                 capability_name,
                                 property_name,
                                 throw_errors=True):
        """Gets a node template capability property."""
        caps = node_template.get_capabilities()
        if caps and capability_name in caps.keys():
            cap = caps[capability_name]
            property = None
            props = cap.get_properties()
            if props and property_name in props.keys():
                property = props[property_name].value
            if property is None and throw_errors:
                ExceptionCollector.appendException(
                    KeyError(_('Property "%(prop)s" was not found in '
                               'capability "%(cap)s" of node template '
                               '"%(ntpl1)s" referenced from node template '
                               '"%(ntpl2)s".') % {'prop': property_name,
                                                  'cap': capability_name,
                                                  'ntpl1': node_template.name,
                                                  'ntpl2': self.context.name}))
            return property
        if throw_errors:
            msg = _('Requirement/Capability "{0}" referenced from '
                    'node template "{1}" was not found in node template'
                    ' "{2}".').format(capability_name,
                                      self.context.name,
                                      node_template.name)
            ExceptionCollector.appendException(KeyError(msg))
        else:
            return None

    def _find_property(self, property_name):
        node_tpl = self._find_node_template(self.args[0])
        if not node_tpl:
            return
        props = node_tpl.get_properties()
        found = [props[property_name]] if property_name in props else []
        if len(found) == 0:
            ExceptionCollector.appendException(
                KeyError(_('Property "%(prop)s" was not found in node '
                           'template "%(ntpl)s".') %
                         {'prop': property_name,
                          'ntpl': node_tpl.name}))
            return None
        return found[0]

    def _find_node_template(self, node_template_name):
        # if the node_template_name has the long format
        if isinstance(node_template_name, dict):
            # get only the node name
            if 'node' in node_template_name:
                node_template_name = node_template_name['node']
            else:
                ExceptionCollector.appendException(
                    ValueError(_(' No node name in the relationship.')))
                return
        if node_template_name == SELF:
            return self.context
        # enable the HOST value in the function
        if node_template_name == HOST:
            node = self._find_host_containing_property()
            if node is None:
                ExceptionCollector.appendException(
                    KeyError(_(
                        "Property '{0}' not found in capability/requirement"
                        " '{1}' referenced from node template {2}").
                        format(self.args[2],
                               self.args[1],
                               self.context.name)))
                return None
            else:
                return node
        if node_template_name == TARGET:
            if not isinstance(self.context.type_definition, RelationshipType):
                ExceptionCollector.appendException(
                    KeyError(_('"TARGET" keyword can only be used in context'
                               ' to "Relationships" target node')))
                return
            return self.context.target
        if node_template_name == SOURCE:
            if not isinstance(self.context.type_definition, RelationshipType):
                ExceptionCollector.appendException(
                    KeyError(_('"SOURCE" keyword can only be used in context'
                               ' to "Relationships" source node')))
                return
            return self.context.source
        if not hasattr(self.tosca_tpl, 'nodetemplates'):
            return
        for node_template in self.tosca_tpl.nodetemplates:
            if node_template.name == node_template_name:
                return node_template
        ExceptionCollector.appendException(
            KeyError(_(
                'Node template "{0}" was not found.'
                ' referenced from node template {1}'
            ).format(node_template_name,
                     self.context.name)))

    def _get_index_value(self, value, index):
        if isinstance(value, list):
            if index < len(value):
                return value[index]
            else:
                ExceptionCollector.appendException(
                    KeyError(_(
                        "Property '{0}' found in capability '{1}'"
                        " referenced from node template {2}"
                        " must have an element with index {3}.").
                        format(self.args[2],
                               self.args[1],
                               self.context.name,
                               index)))
        else:
            ExceptionCollector.appendException(
                KeyError(_(
                    "Property '{0}' found in capability '{1}'"
                    " referenced from node template {2}"
                    " must be a list.").format(self.args[2],
                                               self.args[1],
                                               self.context.name)))

    def _get_attribute_value(self, value, attibute):
        if isinstance(value, dict):
            if attibute in value:
                return value[attibute]
            else:
                ExceptionCollector.appendException(
                    KeyError(_(
                        "Property '{0}' found in capability '{1}'"
                        " referenced from node template {2}"
                        " must have an attribute named {3}.").
                        format(self.args[2],
                               self.args[1],
                               self.context.name,
                               attibute)))
        else:
            ExceptionCollector.appendException(
                KeyError(_(
                    "Property '{0}' found in capability '{1}'"
                    " referenced from node template {2}"
                    " must be a dict.").format(self.args[2],
                                               self.args[1],
                                               self.context.name)))

    # Add this functions similar to get_attribute case
    def _find_host_containing_property(self, node_template_name=SELF):
        node_template = self._find_node_template(node_template_name)
        hosted_on_rel = EntityType.TOSCA_DEF[HOSTED_ON]
        for r in node_template.requirements:
            for requirement, target_name in r.items():
                target_node = self._find_node_template(target_name)
                target_type = target_node.type_definition
                for capability in target_type.get_capabilities_objects():
                    if capability.inherits_from(
                            hosted_on_rel['valid_target_types']):
                        if self._property_exists_in_type(target_type):
                            return target_node
                        # If requirement was not found, look in node
                        # template's capabilities
                        if (len(self.args) > 2 and
                                self._get_capability_property(target_node,
                                                              self.args[1],
                                                              self.args[2],
                                                              False)
                                is not None):
                            return target_node
                        return self._find_host_containing_property(
                            target_name)
        return None

    def _property_exists_in_type(self, type_definition):
        props_def = type_definition.get_properties_def()
        found = [props_def[self.args[1]]] \
            if self.args[1] in props_def else []
        return len(found) == 1

    def result(self):
        if len(self.args) >= 3:
            # First check if there is property with this name
            node_tpl = self._find_node_template(self.args[0])
            props = node_tpl.get_properties() if node_tpl else []
            index = 2
            found = [props[self.args[1]]] if self.args[1] in props else []
            if found:
                property_value = found[0].value
            else:
                index = 3
                # then check the req or caps
                property_value = self._find_req_or_cap_property(self.args[1],
                                                                self.args[2])
            if len(self.args) > index:
                for elem in self.args[index:]:
                    if isinstance(property_value, list):
                        int_elem = int(elem)
                        property_value = self._get_index_value(property_value,
                                                               int_elem)
                    else:
                        property_value = self._get_attribute_value(
                            property_value,
                            elem)
        else:
            property_value = self._find_property(self.args[1]).value
        if isinstance(property_value, Function):
            return property_value.result()
        return get_function(self.tosca_tpl,
                            self.context,
                            property_value)

    @property
    def node_template_name(self):
        return self.args[0]

    @property
    def property_name(self):
        if len(self.args) > 2:
            return self.args[2]
        return self.args[1]

    @property
    def req_or_cap(self):
        if len(self.args) > 2:
            return self.args[1]
        return None


class GetOperationOutput(Function):
    def validate(self):
        if len(self.args) == 4:
            template_name = self.args[0]
            interface_name = self.args[1]
            operation_name = self.args[2]
            template = None
            node_template = self._find_node_template(template_name)
            if node_template:
                template = node_template
            else:
                relationship_template = \
                    self._find_relationship_template(template_name)
                if relationship_template:
                    template = relationship_template
            if not template:
                ExceptionCollector.appendException(
                    KeyError(_(
                        'Node or relationship template "{0}" was not found.'
                    ).format(template_name)))
                return
            if hasattr(template, template.INTERFACES):
                operation = self._find_operation_name(interface_name,
                                                      operation_name,
                                                      template.interfaces)
                if operation:
                    return
            ExceptionCollector.appendException(
                ValueError(_(
                    'Node or relationship template "{0}" has not '
                    'interface "{1}" or operation "{2}".'
                ).format(template_name, interface_name, operation_name)))

        else:
            ExceptionCollector.appendException(
                ValueError(_('Illegal arguments for function "{0}". Expected '
                             'arguments: "template_name","interface_name",'
                             '"operation_name","output_variable_name"'
                             ).format(GET_OPERATION_OUTPUT)))
            return

    def _find_operation_name(self, interface_name, operation_name,
                             interfaces=None):
        if (interface_name == 'Configure' or
           interface_name == 'tosca.interfaces.node.relationship.Configure'):
            if (operation_name in
               StatefulEntityType.
               interfaces_relationship_configure_operations):
                return operation_name
            else:
                ExceptionCollector.appendException(
                    ValueError(_('Enter an operation of Configure interface'
                                 ).format(GET_OPERATION_OUTPUT)))
                return
        elif (interface_name == 'Standard' or
              interface_name == 'tosca.interfaces.node.lifecycle.Standard'):
            if (operation_name in
               StatefulEntityType.interfaces_node_lifecycle_operations):
                return operation_name
            else:
                ExceptionCollector.appendException(
                    ValueError(_('Enter an operation of Standard interface'
                                 ).format(GET_OPERATION_OUTPUT)))
                return
        elif interfaces:
            for interface_obj in interfaces:
                if interface_obj.name == operation_name and \
                        interface_obj.type == interface_name:
                    return operation_name
        else:
            ExceptionCollector.appendException(
                ValueError(_('Enter a valid operation name'
                             ).format(GET_OPERATION_OUTPUT)))
            return

    def _find_node_template(self, node_template_name):
        if node_template_name == TARGET:
            if not isinstance(self.context.type_definition, RelationshipType):
                ExceptionCollector.appendException(
                    KeyError(_('"TARGET" keyword can only be used in context'
                               ' to "Relationships" target node')))
                return
            return self.context.target
        if node_template_name == SOURCE:
            if not isinstance(self.context.type_definition, RelationshipType):
                ExceptionCollector.appendException(
                    KeyError(_('"SOURCE" keyword can only be used in context'
                               ' to "Relationships" source node')))
                return
            return self.context.source
        name = self.context.name \
            if node_template_name == SELF and \
            not isinstance(self.context, list) \
            else node_template_name
        for node_template in self.tosca_tpl.nodetemplates:
            if node_template.name == name:
                return node_template

    def _find_relationship_template(self, relationship_template_name):
        for rel_template in self.tosca_tpl.relationship_templates:
            if rel_template.name == relationship_template_name:
                return rel_template

    def result(self):
        return self


class Concat(Function):
    """Validate the function and provide an instance of the function

    Concatenation of values are supposed to be produced at runtime and
    therefore its the responsibility of the TOSCA engine to implement the
    evaluation of Concat functions.

    Arguments:

    * List of strings that needs to be concatenated

    Example:

      [ 'http://',
        get_attribute: [ server, public_address ],
        ':' ,
        get_attribute: [ server, port ] ]
    """

    def validate(self):
        if len(self.args) < 1:
            ExceptionCollector.appendException(
                ValueError(_('Invalid arguments for function "{0}". Expected '
                             'at least one arguments.').format(CONCAT)))

    def result(self):
        return self


class Token(Function):
    """Validate the function and provide an instance of the function

    The token function is used within a TOSCA service template on a string to
    parse out (tokenize) substrings separated by one or more token characters
    within a larger string.


    Arguments:

    * The composite string that contains one or more substrings separated by
      token characters.
    * The string that contains one or more token characters that separate
      substrings within the composite string.
    * The integer indicates the index of the substring to return from the
      composite string.  Note that the first substring is denoted by using
      the '0' (zero) integer value.

    Example:

     [ get_attribute: [ my_server, data_endpoint, ip_address ], ':', 1 ]

    """

    def validate(self):
        if len(self.args) < 3:
            ExceptionCollector.appendException(
                ValueError(_('Invalid arguments for function "{0}". Expected '
                             'at least three arguments.').format(TOKEN)))
        else:
            if not isinstance(self.args[1], str) or len(self.args[1]) != 1:
                ExceptionCollector.appendException(
                    ValueError(_('Invalid arguments for function "{0}". '
                                 'Expected single char value as second '
                                 'argument.').format(TOKEN)))

            if not isinstance(self.args[2], int):
                ExceptionCollector.appendException(
                    ValueError(_('Invalid arguments for function "{0}". '
                                 'Expected integer value as third '
                                 'argument.').format(TOKEN)))

    def result(self):
        return self


function_mappings = {
    GET_PROPERTY: GetProperty,
    GET_INPUT: GetInput,
    GET_ATTRIBUTE: GetAttribute,
    GET_OPERATION_OUTPUT: GetOperationOutput,
    CONCAT: Concat,
    TOKEN: Token
}


def is_function(function):
    """Returns True if the provided function is a Tosca intrinsic function.

    Examples:

    * "{ get_property: { SELF, port } }"
    * "{ get_input: db_name }"
    * Function instance

    :param function: Function as string or a Function instance.
    :return: True if function is a Tosca intrinsic function, otherwise False.
    """
    if isinstance(function, dict) and len(function) == 1:
        func_name = list(function.keys())[0]
        return func_name in function_mappings
    return isinstance(function, Function)


def get_function(tosca_tpl, node_template, raw_function):
    """Gets a Function instance representing the provided template function.

    If the format provided raw_function format is not relevant for template
    functions or if the function name doesn't exist in function mapping the
    method returns the provided raw_function.

    :param tosca_tpl: The tosca template.
    :param node_template: The node template the function is specified for.
    :param raw_function: The raw function as dict.
    :return: Template function as Function instance or the raw_function if
     parsing was unsuccessful.
    """
    if is_function(raw_function):
        if isinstance(raw_function, dict):
            func_name = list(raw_function.keys())[0]
            if func_name in function_mappings:
                func = function_mappings[func_name]
                func_args = list(raw_function.values())[0]
                if not isinstance(func_args, list):
                    func_args = [func_args]
                return func(tosca_tpl, node_template, func_name, func_args)
    return raw_function



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from toscaparser.tests.base import TestCase
from toscaparser.tests import utils
from toscaparser.tosca_template import ToscaTemplate


class ToscaNFVTemplateTest(TestCase):

    '''TOSCA NFV template.'''
    tosca_tpl = utils.get_sample_test_path(
        "data/extensions/tosca_helloworld_nfv.yaml")
    tosca = ToscaTemplate(tosca_tpl)

    def test_version(self):
        self.assertEqual(self.tosca.version,
                         "tosca_simple_profile_for_nfv_1_0_0")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from toscaparser.common import exception
from toscaparser.tests.base import TestCase
from toscaparser.utils.gettextutils import _


class ExceptionTest(TestCase):

    def setUp(self):
        super(TestCase, self).setUp()
        exception.TOSCAException.set_fatal_format_exception(False)

    def test_message(self):
        ex = exception.MissingRequiredFieldError(what='Template',
                                                 required='type')
        self.assertEqual(_('Template is missing required field "type".'),
                         ex.__str__())

    def test_set_flag(self):
        exception.TOSCAException.set_fatal_format_exception('True')
        self.assertFalse(
            exception.TOSCAException._FATAL_EXCEPTION_FORMAT_ERRORS)

    def test_format_error(self):
        ex = exception.UnknownFieldError(what='Template')
        self.assertEqual(_('An unknown exception occurred.'), ex.__str__(),)
        self.assertRaises(KeyError, self._formate_exception)

    def _formate_exception(self):
        exception.UnknownFieldError.set_fatal_format_exception(True)
        raise exception.UnknownFieldError(what='Template')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from toscaparser.common.exception import ExceptionCollector
from toscaparser.common.exception import UnknownFieldError
from toscaparser.elements.statefulentitytype import StatefulEntityType


class RelationshipType(StatefulEntityType):
    '''TOSCA built-in relationship type.'''
    SECTIONS = (DERIVED_FROM, VALID_TARGET_TYPES, INTERFACES,
                ATTRIBUTES, PROPERTIES, DESCRIPTION, VERSION,
                CREDENTIAL) = ('derived_from', 'valid_target_types',
                               'interfaces', 'attributes', 'properties',
                               'description', 'version', 'credential')

    def __init__(self, type, capability_name=None, custom_def=None):
        super(RelationshipType, self).__init__(type, self.RELATIONSHIP_PREFIX,
                                               custom_def)
        self.capability_name = capability_name
        self.custom_def = custom_def
        self._validate_keys()

    @property
    def parent_type(self):
        '''Return a relationship this reletionship is derived from.'''
        prel = self.derived_from(self.defs)
        if prel:
            return RelationshipType(prel, custom_def=self.custom_def)

    @property
    def interfaces(self):
        interfaces = self.get_value(self.INTERFACES)

        if self.parent_type is not None:
            if self.parent_type.interfaces is not None:
                import copy
                parent_interfaces = copy.deepcopy(self.parent_type.interfaces)

                if parent_interfaces:
                    if interfaces:
                        parent_interfaces.update(interfaces)
                    interfaces = parent_interfaces
        return interfaces

    @property
    def valid_target_types(self):
        return self.entity_value(self.defs, 'valid_target_types')

    def _validate_keys(self):
        for key in self.defs.keys():
            if key not in self.SECTIONS:
                ExceptionCollector.appendException(
                    UnknownFieldError(what='Relationshiptype "%s"' % self.type,
                                      field=key))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import urllib

from toscaparser.common import exception
from toscaparser.imports import ImportsLoader
from toscaparser.nodetemplate import NodeTemplate
from toscaparser.parameters import Input
from toscaparser.parameters import Output
from toscaparser.policy import Policy
from toscaparser.relationship_template import RelationshipTemplate
from toscaparser.repositories import Repository
from toscaparser.reservation import Reservation
from toscaparser.tests.base import MockTestClass
from toscaparser.tests.base import TestCase
from toscaparser.tests import utils
from toscaparser.topology_template import TopologyTemplate
from toscaparser.tosca_template import ToscaTemplate
from toscaparser.triggers import Triggers
from toscaparser.utils.gettextutils import _
import toscaparser.utils.yamlparser


class ToscaTemplateValidationTest(TestCase):

    def test_well_defined_template(self):
        tpl_path = utils.get_sample_test_path(
            "data/tosca_single_instance_wordpress.yaml")
        params = {'db_name': 'my_wordpress', 'db_user': 'my_db_user',
                  'db_root_pwd': '12345678'}
        self.assertIsNotNone(ToscaTemplate(tpl_path, params))

    def test_custom_interface_allowed(self):
        tpl_path = utils.get_sample_test_path(
            "data/interfaces/test_custom_interface_in_template.yaml")
        self.assertIsNotNone(ToscaTemplate(tpl_path))

    def test_custom_interface_invalid_operation(self):
        tpl_path = utils.get_sample_test_path(
            "data/interfaces/test_custom_interface_invalid_operation.yaml")
        self.assertRaises(exception.ValidationError,
                          ToscaTemplate, tpl_path)
        exception.ExceptionCollector.assertExceptionMessage(
            exception.UnknownFieldError,
            _('"interfaces" of template "customInterfaceTest" '
              'contains unknown field "CustomOp4". '
              'Refer to the definition to verify valid values.'))

    # Since 'operations' Keyname was defined in TOSCA1.3,
    # test whether operation_definitions defined
    # under 'operations' can be used.
    def test_custom_interface_operations(self):
        tpl_path = utils.get_sample_test_path(
            "data/interfaces/test_custom_interface_operations.yaml")
        tosca = ToscaTemplate(tpl_path)
        op_names = tosca.tpl['interface_types'][
            'tosca.interfaces.CustomInterface']['operations'].keys()
        self.assertEqual('CustomOp', list(op_names)[0])
        op_names = tosca.topology_template.custom_defs[
            'tosca.interfaces.CustomInterface']['operations'].keys()
        self.assertEqual('CustomOp', list(op_names)[0])
        op_names = tosca.topology_template.nodetemplates[0].templates[
            'customInterfaceTest']['interfaces']['CustomInterface'][
                'operations'].keys()
        self.assertEqual('CustomOp', list(op_names)[0])

    # Since 'notifications' Keyname was defined in TOSCA1.3,
    # test whether notification_definitions defined
    # under 'notifications' can be used.
    def test_custom_interface_notifications(self):
        tpl_path = utils.get_sample_test_path(
            "data/interfaces/test_custom_interface_notifications.yaml")
        tosca = ToscaTemplate(tpl_path)
        no_names = tosca.tpl['interface_types'][
            'tosca.interfaces.CustomInterface']['notifications'].keys()
        self.assertEqual('CustomNo', list(no_names)[0])
        no_names = tosca.topology_template.custom_defs[
            'tosca.interfaces.CustomInterface']['notifications'].keys()
        self.assertEqual('CustomNo', list(no_names)[0])
        no_names = tosca.topology_template.nodetemplates[0].templates[
            'customInterfaceTest']['interfaces']['CustomInterface'][
                'notifications'].keys()
        self.assertEqual('CustomNo', list(no_names)[0])

    def test_first_level_sections(self):
        tpl_path = utils.get_sample_test_path(
            "data/test_tosca_top_level_error1.yaml")
        self.assertRaises(exception.ValidationError, ToscaTemplate, tpl_path)
        exception.ExceptionCollector.assertExceptionMessage(
            exception.MissingRequiredFieldError,
            _('Template is missing required field '
              '"tosca_definitions_version".'))

        tpl_path = utils.get_sample_test_path(
            "data/test_tosca_top_level_error2.yaml")
        self.assertRaises(exception.ValidationError, ToscaTemplate, tpl_path)
        exception.ExceptionCollector.assertExceptionMessage(
            exception.UnknownFieldError,
            _('Template contains unknown field "node_template". Refer to the '
              'definition to verify valid values.'))

    def test_template_with_imports_validation(self):
        tpl_path = utils.get_sample_test_path(
            "data/tosca_imports_validation.yaml")
        self.assertRaises(exception.ValidationError, ToscaTemplate, tpl_path)
        exception.ExceptionCollector.assertExceptionMessage(
            exception.UnknownFieldError,
            _('Template custom_types/imported_sample.yaml contains unknown '
              'field "descriptions". Refer to the definition'
              ' to verify valid values.'))
        exception.ExceptionCollector.assertExceptionMessage(
            exception.UnknownFieldError,
            _('Template custom_types/imported_sample.yaml contains unknown '
              'field "node_typess". Refer to the definition to '
              'verify valid values.'))
        exception.ExceptionCollector.assertExceptionMessage(
            exception.UnknownFieldError,
            _('Template custom_types/imported_sample.yaml contains unknown '
              'field "tosca1_definitions_version". Refer to the definition'
              ' to verify valid values.'))
        versions = '", "'.join(ToscaTemplate.VALID_TEMPLATE_VERSIONS)
        exception.ExceptionCollector.assertExceptionMessage(
            exception.InvalidTemplateVersion,
            _('The template version "tosca_simple_yaml_1_10 in '
              'custom_types/imported_sample.yaml" is invalid. '
              'Valid versions are "%s".') % versions)
        exception.ExceptionCollector.assertExceptionMessage(
            exception.UnknownFieldError,
            _('Template custom_types/imported_sample.yaml contains unknown '
              'field "policy_types1". Refer to the definition to '
              'verify valid values.'))
        exception.ExceptionCollector.assertExceptionMessage(
            exception.UnknownFieldError,
            _('Nodetype"tosca.nodes.SoftwareComponent.Logstash" contains '
              'unknown field "capabilities1". Refer to the definition '
              'to verify valid values.'))
        exception.ExceptionCollector.assertExceptionMessage(
            exception.UnknownFieldError,
            _('Policy "mycompany.mytypes.myScalingPolicy" contains unknown '
              'field "derived1_from". Refer to the definition to '
              'verify valid values.'))
        exception.ExceptionCollector.assertExceptionMessage(
            exception.UnknownFieldError,
            _('Relationshiptype "test.relation.connects" contains unknown '
              'field "derived_from4". Refer to the definition to '
              'verify valid values.'))

    def test_getoperation_IncorrectValue(self):
        # test case 1
        tpl_snippet = '''
        node_templates:
             front_end:
               type: tosca.nodes.Compute
               interfaces:
                 Standard:
                   create:
                     implementation: scripts/frontend/create.sh
                   configure:
                     implementation: scripts/frontend/configure.sh
                     inputs:
                       data_dir: {get_operation_output: [front_end,Standard1,
                                                            create,data_dir]}
        '''
        tpl = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        err = self.assertRaises(ValueError,
                                TopologyTemplate, tpl, None)
        expectedmessage = _('Node or relationship template "front_end" '
                            'has not interface "Standard1" '
                            'or operation "create".')
        self.assertEqual(expectedmessage, err.__str__())
        # test case 2
        tpl_snippet2 = '''
        node_templates:
             front_end:
               type: tosca.nodes.Compute
               interfaces:
                 Standard:
                   create:
                     implementation: scripts/frontend/create.sh
                   configure:
                     implementation: scripts/frontend/configure.sh
                     inputs:
                       data_dir: {get_operation_output: [front_end1,Standard,
                                                            create,data_dir]}
        '''
        tpl2 = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet2))
        err2 = self.assertRaises(KeyError,
                                 TopologyTemplate, tpl2, None)
        expectedmessage2 = _('\'Node or relationship template "front_end1" '
                             'was not found.\'')
        self.assertEqual(expectedmessage2, err2.__str__())
        # test case 3
        tpl_snippet3 = '''
        node_templates:
             front_end:
               type: tosca.nodes.Compute
               interfaces:
                 Standard:
                   create:
                     implementation: scripts/frontend/create.sh
                   configure:
                     implementation: scripts/frontend/configure.sh
                     inputs:
                       data_dir: {get_operation_output: [front_end,Standard,
                                                      get_target,data_dir]}
        '''
        tpl3 = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet3))
        err3 = self.assertRaises(ValueError,
                                 TopologyTemplate, tpl3, None)
        expectedmessage3 = _('Enter an operation of Standard interface')
        self.assertEqual(expectedmessage3, err3.__str__())
        # test case 4
        tpl_snippet4 = '''
        node_templates:
             front_end:
               type: tosca.nodes.Compute
               interfaces:
                 Standard:
                   create:
                     implementation: scripts/frontend/create.sh
                   configure:
                     implementation: scripts/frontend/configure.sh
                     inputs:
                       data_dir: {get_operation_output: [front_end,Configure,
                                                        create,data_dir]}
        '''
        tpl4 = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet4))
        err4 = self.assertRaises(ValueError,
                                 TopologyTemplate, tpl4, None)
        expectedmessage4 = _('Enter an operation of Configure interface')
        self.assertEqual(expectedmessage4, err4.__str__())
        # test case 5
        tpl_snippet5 = '''
        node_templates:
             front_end:
               type: tosca.nodes.Compute
               interfaces:
                 Standard:
                   create:
                     implementation: scripts/frontend/create.sh
                   configure:
                     implementation: scripts/frontend/configure.sh
                     inputs:
                       data_dir: {get_operation_output: [front_end,Standard,
                                                                    create]}
        '''
        tpl5 = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet5))
        err5 = self.assertRaises(ValueError,
                                 TopologyTemplate, tpl5, None)
        expectedmessage5 = _('Illegal arguments for function'
                             ' "get_operation_output".'
                             ' Expected arguments: "template_name",'
                             '"interface_name",'
                             '"operation_name","output_variable_name"')
        self.assertEqual(expectedmessage5, err5.__str__())

    def test_unsupported_type(self):
        tpl_snippet = '''
        node_templates:
          invalid_type:
            type: tosca.test.invalidtype
            properties:
              size: { get_input: storage_size }
              snapshot_id: { get_input: storage_snapshot_id }
        '''
        tpl = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        err = self.assertRaises(exception.UnsupportedTypeError,
                                TopologyTemplate, tpl, None)
        expectedmessage = _('Type "tosca.test.invalidtype" is valid'
                            ' TOSCA type but not supported at this time.')
        self.assertEqual(expectedmessage, err.__str__())

    def test_inputs(self):
        tpl_snippet1 = '''
        inputs:
          cpus:
            type: integer
            description: Number of CPUs for the server.
            constraint:
              - valid_values: [ 1, 2, 4 ]
            required: yes
            status: supported
        '''
        tpl_snippet2 = '''
        inputs:
          cpus:
            type: integer
            description: Number of CPUs for the server.
            constraints:
              - valid_values: [ 1, 2, 4 ]
            required: yes
            status: supported
        '''
        tpl_snippet3 = '''
        inputs:
          some_list:
            type: list
            description: List of items
            entry_schema:
              type: string
            default: []
        '''
        tpl_snippet4 = '''
        inputs:
          some_list:
            type: list
            description: List of items
            key_schema:
              type: string
            default: []
        '''
        inputs1 = (toscaparser.utils.yamlparser.
                   simple_parse(tpl_snippet1)['inputs'])
        name1, attrs1 = list(inputs1.items())[0]
        inputs2 = (toscaparser.utils.yamlparser.
                   simple_parse(tpl_snippet2)['inputs'])
        name2, attrs2 = list(inputs2.items())[0]
        try:
            Input(name1, attrs1)
        except Exception as err:
            self.assertEqual(_('Input "cpus" contains unknown field '
                               '"constraint". Refer to the definition to '
                               'verify valid values.'),
                             err.__str__())
        input2 = Input(name2, attrs2)
        self.assertTrue(input2.required)
        toscaparser.utils.yamlparser.simple_parse(tpl_snippet3)['inputs']
        input4 = (toscaparser.utils.yamlparser.
                  simple_parse(tpl_snippet4)['inputs'])
        self.assertEqual('string', input4['some_list']['key_schema']['type'])

    def _imports_content_test(self, tpl_snippet, path, custom_type_def):
        imports = (toscaparser.utils.yamlparser.
                   simple_parse(tpl_snippet)['imports'])
        loader = ImportsLoader(imports, path, custom_type_def)
        return loader.get_custom_defs()

    def test_imports_without_templates(self):
        tpl_snippet = '''
        imports:
          # omitted here for brevity
        '''
        path = utils.get_sample_test_path("data/tosca_elk.yaml")
        errormsg = _('"imports" keyname is defined without including '
                     'templates.')
        err = self.assertRaises(exception.ValidationError,
                                self._imports_content_test,
                                tpl_snippet,
                                path,
                                "node_types")
        self.assertEqual(errormsg, err.__str__())

    def test_imports_with_name_without_templates(self):
        tpl_snippet = '''
        imports:
          - some_definitions:
        '''
        path = utils.get_sample_test_path("data/tosca_elk.yaml")
        errormsg = _('A template file name is not provided with import '
                     'definition "some_definitions".')
        err = self.assertRaises(exception.ValidationError,
                                self._imports_content_test,
                                tpl_snippet, path, None)
        self.assertEqual(errormsg, err.__str__())

    @mock.patch.object(urllib.request, 'urlopen')
    def test_imports_without_import_name(self, mock_urlopen):
        tpl_snippet = '''
        imports:
          - custom_types/paypalpizzastore_nodejs_app.yaml
          - https://example.com/custom_types/wordpress.yaml
        '''
        mock_path = "https://example.com/custom_types/wordpress.yaml"

        mockclass = MockTestClass()
        mockclass.comp_urldict = {mock_path: utils.get_sample_test_path(
            "data/custom_types/wordpress.yaml")}
        mock_urlopen.side_effect = mockclass.mock_urlopen_method
        path = utils.get_sample_test_path("data/tosca_elk.yaml")
        custom_defs = self._imports_content_test(tpl_snippet,
                                                 path,
                                                 "node_types")
        self.assertTrue(custom_defs)

    @mock.patch.object(urllib.request, 'urlopen')
    def test_imports_wth_import_name(self, mock_urlopen):
        tpl_snippet = '''
        imports:
          - some_definitions: custom_types/paypalpizzastore_nodejs_app.yaml
          - more_definitions:
              file: 'https://example.com/custom_types/wordpress.yaml'
              namespace_prefix: single_instance_wordpress
        '''
        mock_path = "https://example.com/custom_types/wordpress.yaml"

        mockclass = MockTestClass()
        mockclass.comp_urldict = {mock_path: utils.get_sample_test_path(
            "data/custom_types/wordpress.yaml")}
        mock_urlopen.side_effect = mockclass.mock_urlopen_method
        path = utils.get_sample_test_path("data/tosca_elk.yaml")
        custom_defs = self._imports_content_test(tpl_snippet,
                                                 path,
                                                 "node_types")
        self.assertTrue(custom_defs.get("single_instance_wordpress.tosca."
                                        "nodes.WebApplication.WordPress"))

    def test_imports_wth_namespace_prefix(self):
        tpl_snippet = '''
        imports:
          - more_definitions:
              file: custom_types/nested_rsyslog.yaml
              namespace_prefix: testprefix
        '''
        path = utils.get_sample_test_path("data/tosca_elk.yaml")
        custom_defs = self._imports_content_test(tpl_snippet,
                                                 path,
                                                 "node_types")
        self.assertTrue(custom_defs.get("testprefix.Rsyslog"))

    def test_imports_with_no_main_template(self):
        tpl_snippet = '''
        imports:
          - some_definitions: https://raw.githubusercontent.com/openstack/\
tosca-parser/master/toscaparser/tests/data/custom_types/wordpress.yaml
          - some_definitions:
              file: my_defns/my_typesdefs_n.yaml
        '''
        errormsg = _('Input tosca template is not provided.')
        err = self.assertRaises(exception.ValidationError,
                                self._imports_content_test,
                                tpl_snippet, None, None)
        self.assertEqual(errormsg, err.__str__())

    @mock.patch.object(urllib.request, 'urlopen')
    def test_imports_duplicate_name(self, mock_urlopen):
        tpl_snippet = '''
        imports:
          - some_definitions: https://example.com/custom_types/wordpress.yaml
          - some_definitions:
              file: my_defns/my_typesdefs_n.yaml
        '''
        mock_path = "https://example.com/custom_types/wordpress.yaml"

        mockclass = MockTestClass()
        mockclass.comp_urldict = {mock_path: utils.get_sample_test_path(
            "data/custom_types/wordpress.yaml")}
        mock_urlopen.side_effect = mockclass.mock_urlopen_method
        errormsg = _('Duplicate import name "some_definitions" was found.')
        path = utils.get_sample_test_path("data/tosca_elk.yaml")
        err = self.assertRaises(exception.ValidationError,
                                self._imports_content_test,
                                tpl_snippet, path, None)
        self.assertEqual(errormsg, err.__str__())

    def test_imports_missing_req_field_in_def(self):
        tpl_snippet = '''
        imports:
          - more_definitions:
              file1: my_defns/my_typesdefs_n.yaml
              repository: my_company_repo
              namespace_uri: http://mycompany.com/ns/tosca/2.0
              namespace_prefix: mycompany
        '''
        errormsg = _('Import of template "more_definitions" is missing '
                     'required field "file".')
        path = utils.get_sample_test_path("data/tosca_elk.yaml")
        err = self.assertRaises(exception.MissingRequiredFieldError,
                                self._imports_content_test,
                                tpl_snippet, path, None)
        self.assertEqual(errormsg, err.__str__())

    @mock.patch.object(urllib.request, 'urlopen')
    @mock.patch.object(ToscaTemplate, '_tpl_imports')
    def test_imports_file_with_uri(self, mock_tpl_imports, mock_urlopen):
        tpl_snippet = '''
        imports:
          - more_definitions:
              file: https://example.com/custom_types/wordpress.yaml
        '''
        path = ('https://example.com/tosca_single_instance_wordpress_'
                'with_url_import.yaml')
        mock_path = "https://example.com/custom_types/wordpress.yaml"

        import_file_path = utils.get_sample_test_path(
            "data/custom_types/wordpress.yaml")
        mockclass = MockTestClass()
        mockclass.comp_urldict = {mock_path: import_file_path}
        mock_urlopen.side_effect = mockclass.mock_urlopen_method
        mock_tpl_imports.return_value = [import_file_path]
        custom_defs = self._imports_content_test(tpl_snippet,
                                                 path,
                                                 "node_types")
        self.assertTrue(custom_defs.get("tosca.nodes."
                                        "WebApplication.WordPress"))

    @mock.patch.object(urllib.request, 'urlopen')
    def test_imports_file_namespace_fields(self, mock_urlopen):
        tpl_snippet = '''
        imports:
          - more_definitions:
             file: https://example.com/custom_types/wordpress.yaml
             namespace_prefix: mycompany
             namespace_uri: http://docs.oasis-open.org/tosca/ns/simple/yaml/1.0
        '''
        mock_path = "https://example.com/custom_types/wordpress.yaml"

        mockclass = MockTestClass()
        mockclass.comp_urldict = {mock_path: utils.get_sample_test_path(
            "data/custom_types/wordpress.yaml")}
        mock_urlopen.side_effect = mockclass.mock_urlopen_method
        path = utils.get_sample_test_path("data/tosca_elk.yaml")
        custom_defs = self._imports_content_test(tpl_snippet,
                                                 path,
                                                 "node_types")
        self.assertTrue(custom_defs.get("mycompany.tosca.nodes."
                                        "WebApplication.WordPress"))

    @mock.patch.object(urllib.request, 'urlopen')
    def test_imports_with_local_defs(self, mock_urlopen):
        """Compare custom types on local and remote."""

        ctypes = {
            "remote": "https://example.com/custom_types/wordpress.yaml",
            "local": "../data/wordpress.yaml"}
        mock_path = "https://example.com/custom_types/wordpress.yaml"

        mockclass = MockTestClass()
        mockclass.comp_urldict = {mock_path: utils.get_sample_test_path(
            "data/custom_types/wordpress.yaml")}
        mock_urlopen.side_effect = mockclass.mock_urlopen_method

        tpl_snippet = '''
        imports:
        - {}
        '''.format(ctypes["remote"])
        local_defs = {ctypes["remote"]: ctypes["local"]}

        path = utils.get_sample_test_path("data/tosca_elk.yaml")
        imports = (toscaparser.utils.yamlparser.
                   simple_parse(tpl_snippet)['imports'])
        ld1 = ImportsLoader(imports, path, "node_types")
        ld2 = ImportsLoader(imports, path, "node_types", local_defs)
        self.assertEqual(ld1.get_custom_defs(), ld2.get_custom_defs())

    def test_imports_file_with_suffix_yml(self):
        tpl_snippet = '''
        imports:
        - custom_types/wordpress.yml
        '''
        path = utils.get_sample_test_path("data/tosca_elk.yaml")
        custom_defs = self._imports_content_test(tpl_snippet,
                                                 path,
                                                 "node_types")
        self.assertTrue(custom_defs.get("tosca.nodes."
                                        "WebApplication.WordPress"))

    def test_import_error_file_uri(self):
        tpl_snippet = '''
        imports:
          - more_definitions:
             file: mycompany.com/ns/tosca/2.0/toscaparser/tests/data\
/tosca_elk.yaml
             namespace_prefix: mycompany
             namespace_uri: http://docs.oasis-open.org/tosca/ns/simple/yaml/1.0
        '''
        path = utils.get_sample_test_path("data/tosca_elk.yaml")
        self.assertRaises(ImportError,
                          self._imports_content_test,
                          tpl_snippet, path, None)

    def test_import_single_line_error(self):
        tpl_snippet = '''
        imports:
          - some_definitions: abc.com/tests/data/tosca_elk.yaml
        '''
        errormsg = _('Import "abc.com/tests/data/tosca_elk.yaml" is not '
                     'valid.')
        path = utils.get_sample_test_path("data/tosca_elk.yaml")
        err = self.assertRaises(ImportError,
                                self._imports_content_test,
                                tpl_snippet, path, None)
        self.assertEqual(errormsg, err.__str__())

    def test_outputs(self):
        tpl_snippet = '''
        outputs:
          server_address:
            description: IP address of server instance.
            values: { get_property: [server, private_address] }
        '''
        outputs = (toscaparser.utils.yamlparser.
                   simple_parse(tpl_snippet)['outputs'])
        name, attrs = list(outputs.items())[0]
        output = Output(name, attrs)
        try:
            output.validate()
        except Exception as err:
            self.assertIsInstance(err, exception.MissingRequiredFieldError)
            self.assertEqual(_('Output "server_address" is missing required '
                               'field "value".'), err.__str__())

        tpl_snippet = '''
        outputs:
          server_address:
            descriptions: IP address of server instance.
            value: { get_property: [server, private_address] }
        '''
        outputs = (toscaparser.utils.yamlparser.
                   simple_parse(tpl_snippet)['outputs'])
        name, attrs = list(outputs.items())[0]
        output = Output(name, attrs)
        try:
            output.validate()
        except Exception as err:
            self.assertIsInstance(err, exception.UnknownFieldError)
            self.assertEqual(_('Output "server_address" contains unknown '
                               'field "descriptions". Refer to the definition '
                               'to verify valid values.'),
                             err.__str__())

    def _repo_content(self, path):
        repositories = path['repositories']
        reposit = []
        for name, val in repositories.items():
            reposits = Repository(name, val)
            reposit.append(reposits)
        return reposit

    def test_repositories(self):
        tpl_snippet = '''
        repositories:
           repo_code0: https://raw.githubusercontent.com/nandinivemula/intern
           repo_code1:
              description: My project's code Repository in github usercontent.
              url: https://github.com/nandinivemula/intern
              credential:
                 user: nandini
                 password: tcs@12345
           repo_code2:
              description: My Project's code Repository in github.
              url: https://github.com/nandinivemula/intern
              credential:
                 user: xyzw
                 password: xyz@123
        '''
        tpl = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        repoobject = self._repo_content(tpl)
        actualrepo_names = []
        for repo in repoobject:
            repos = repo.name
            actualrepo_names.append(repos)
        reposname = list(tpl.values())
        reposnames = reposname[0]
        expected_reponames = list(reposnames.keys())
        self.assertEqual(expected_reponames, actualrepo_names)

    def test_repositories_with_missing_required_field(self):
        tpl_snippet = '''
        repositories:
           repo_code0: https://raw.githubusercontent.com/nandinivemula/intern
           repo_code1:
              description: My project's code Repository in github usercontent.
              credential:
                 user: nandini
                 password: tcs@12345
           repo_code2:
              description: My Project's code Repository in github.
              url: https://github.com/nandinivemula/intern
              credential:
                 user: xyzw
                 password: xyz@123
        '''
        tpl = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        err = self.assertRaises(exception.MissingRequiredFieldError,
                                self._repo_content, tpl)
        expectedmessage = _('Repository "repo_code1" is missing '
                            'required field "url".')
        self.assertEqual(expectedmessage, err.__str__())

    def test_repositories_with_unknown_field(self):
        tpl_snippet = '''
        repositories:
           repo_code0: https://raw.githubusercontent.com/nandinivemula/intern
           repo_code1:
              description: My project's code Repository in github usercontent.
              url: https://github.com/nandinivemula/intern
              credential:
                 user: nandini
                 password: tcs@12345
           repo_code2:
              descripton: My Project's code Repository in github.
              url: https://github.com/nandinivemula/intern
              credential:
                 user: xyzw
                 password: xyz@123
        '''
        tpl = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        err = self.assertRaises(exception.UnknownFieldError,
                                self._repo_content, tpl)
        expectedmessage = _('repositories "repo_code2" contains unknown field'
                            ' "descripton". Refer to the definition to verify'
                            ' valid values.')
        self.assertEqual(expectedmessage, err.__str__())

    def test_repositories_with_invalid_url(self):
        tpl_snippet = '''
        repositories:
           repo_code0: https://raw.githubusercontent.com/nandinivemula/intern
           repo_code1:
              description: My project's code Repository in github usercontent.
              url: h
              credential:
                 user: nandini
                 password: tcs@12345
           repo_code2:
              description: My Project's code Repository in github.
              url: https://github.com/nandinivemula/intern
              credential:
                 user: xyzw
                 password: xyz@123
        '''
        tpl = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        err = self.assertRaises(exception.URLException,
                                self._repo_content, tpl)
        expectedmessage = _('repsositories "repo_code1" Invalid Url')
        self.assertEqual(expectedmessage, err.__str__())

    def test_groups(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            requirements:
              - log_endpoint:
                  capability: log_endpoint

          mysql_dbms:
            type: tosca.nodes.DBMS
            properties:
              root_password: aaa
              port: 3376
            requirements:
            - host: server

        groups:
          webserver_group:
            type: tosca.groups.Root
            members: [ server, mysql_dbms ]
        '''
        tpl = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        TopologyTemplate(tpl, None)

    def test_groups_with_missing_required_field(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            requirements:
              - log_endpoint:
                  capability: log_endpoint

          mysql_dbms:
            type: tosca.nodes.DBMS
            properties:
              root_password: aaa
              port: 3376
            requirements:
            - host: server

        groups:
          webserver_group:
              members: ['server', 'mysql_dbms']
        '''
        tpl = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        err = self.assertRaises(exception.MissingRequiredFieldError,
                                TopologyTemplate, tpl, None)
        expectedmessage = _('Template "webserver_group" is missing '
                            'required field "type".')
        self.assertEqual(expectedmessage, err.__str__())

    def test_groups_with_unknown_target(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            requirements:
              - log_endpoint:
                  capability: log_endpoint

          mysql_dbms:
            type: tosca.nodes.DBMS
            properties:
              root_password: aaa
              port: 3376
            requirements:
            - host: server

        groups:
          webserver_group:
            type: tosca.groups.Root
            members: [ serv, mysql_dbms ]
        '''
        tpl = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        expectedmessage = _('"Target member "serv" is not found in '
                            'node_templates"')
        err = self.assertRaises(exception.InvalidGroupTargetException,
                                TopologyTemplate, tpl, None)
        self.assertEqual(expectedmessage, err.__str__())

    def test_groups_with_repeated_targets(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            requirements:
              - log_endpoint:
                  capability: log_endpoint

          mysql_dbms:
            type: tosca.nodes.DBMS
            properties:
              root_password: aaa
              port: 3376
            requirements:
            - host: server

        groups:
          webserver_group:
            type: tosca.groups.Root
            members: [ server, server, mysql_dbms ]
        '''
        tpl = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        expectedmessage = _('"Member nodes '
                            '"[\'server\', \'server\', \'mysql_dbms\']" '
                            'should be >= 1 and not repeated"')
        err = self.assertRaises(exception.InvalidGroupTargetException,
                                TopologyTemplate, tpl, None)
        self.assertEqual(expectedmessage, err.__str__())

    def test_groups_with_only_one_target(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            requirements:
              - log_endpoint:
                  capability: log_endpoint

          mysql_dbms:
            type: tosca.nodes.DBMS
            properties:
              root_password: aaa
              port: 3376
            requirements:
            - host: server

        groups:
          webserver_group:
            type: tosca.groups.Root
            members: []
        '''
        tpl = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        expectedmessage = _('"Member nodes "[]" should be >= 1 '
                            'and not repeated"')
        err = self.assertRaises(exception.InvalidGroupTargetException,
                                TopologyTemplate, tpl, None)
        self.assertEqual(expectedmessage, err.__str__())

    def test_groups_with_attributes(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            requirements:
              - log_endpoint:
                  capability: log_endpoint

        groups:
          webserver_group:
            type: tosca.groups.Root
            attributes:
              cpu_usage:
                description: 'Current CPU usage of the node'
                value: 75
              members: [ server ]
        '''
        tpl = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        groups = TopologyTemplate(tpl, None).groups
        self.assertEqual('Current CPU usage of the node',
                         groups[0].entity_tpl['attributes']
                         ['cpu_usage']['description'])
        self.assertEqual(75, groups[0].entity_tpl['attributes']
                         ['cpu_usage']['value'])

    def _custom_types(self):
        custom_types = {}
        def_file = utils.get_sample_test_path(
            "data/custom_types/wordpress.yaml")
        custom_type = toscaparser.utils.yamlparser.load_yaml(def_file)
        node_types = custom_type['node_types']
        for name in node_types:
            defintion = node_types[name]
            custom_types[name] = defintion
        return custom_types

    def _custom_types_policy(self):
        custom_types = {}
        def_file = utils.get_sample_test_path(
            "data/custom_types/custom_policy.yaml")
        custom_type = toscaparser.utils.yamlparser.load_yaml(def_file)
        policy_types = custom_type['policy_types']
        for name in policy_types:
            defintion = policy_types[name]
            custom_types[name] = defintion
        return custom_types

    def _single_node_template_content_test(self, tpl_snippet):
        nodetemplates = (toscaparser.utils.yamlparser.
                         simple_ordered_parse(tpl_snippet))['node_templates']
        name = list(nodetemplates.keys())[0]
        nodetemplate = NodeTemplate(name, nodetemplates,
                                    self._custom_types())
        nodetemplate.validate()
        nodetemplate.requirements
        nodetemplate.get_capabilities_objects()
        nodetemplate.get_properties_objects()
        nodetemplate.interfaces

    def test_node_templates(self):
        tpl_snippet = '''
        node_templates:
          server:
            capabilities:
              host:
                properties:
                  disk_size: 10
                  num_cpus: 4
                  mem_size: 4096
              os:
                properties:
                  architecture: x86_64
                  type: Linux
                  distribution: Fedora
                  version: 18.0
        '''
        expectedmessage = _('Template "server" is missing required field '
                            '"type".')
        err = self.assertRaises(
            exception.MissingRequiredFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_with_wrong_properties_keyname(self):
        """Node template keyname 'properties' given as 'propertiessss'."""
        tpl_snippet = '''
        node_templates:
          mysql_dbms:
            type: tosca.nodes.DBMS
            propertiessss:
              root_password: aaa
              port: 3376
        '''
        expectedmessage = _('Node template "mysql_dbms" contains unknown '
                            'field "propertiessss". Refer to the definition '
                            'to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_with_wrong_requirements_keyname(self):
        """Node template keyname 'requirements' given as 'requirement'."""
        tpl_snippet = '''
        node_templates:
          mysql_dbms:
            type: tosca.nodes.DBMS
            properties:
              root_password: aaa
              port: 3376
            requirement:
              - host: server
        '''
        expectedmessage = _('Node template "mysql_dbms" contains unknown '
                            'field "requirement". Refer to the definition to '
                            'verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_with_wrong_interfaces_keyname(self):
        """Node template keyname 'interfaces' given as 'interfac'."""
        tpl_snippet = '''
        node_templates:
          mysql_dbms:
            type: tosca.nodes.DBMS
            properties:
              root_password: aaa
              port: 3376
            requirements:
              - host: server
            interfac:
              Standard:
                configure: mysql_database_configure.sh
        '''
        expectedmessage = _('Node template "mysql_dbms" contains unknown '
                            'field "interfac". Refer to the definition to '
                            'verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_with_wrong_capabilities_keyname(self):
        """Node template keyname 'capabilities' given as 'capabilitiis'."""
        tpl_snippet = '''
        node_templates:
          mysql_database:
            type: tosca.nodes.Database
            properties:
              db_name: { get_input: db_name }
              db_user: { get_input: db_user }
              db_password: { get_input: db_pwd }
            capabilitiis:
              database_endpoint:
                properties:
                  port: { get_input: db_port }
        '''
        expectedmessage = _('Node template "mysql_database" contains unknown '
                            'field "capabilitiis". Refer to the definition to '
                            'verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_with_wrong_artifacts_keyname(self):
        """Node template keyname 'artifacts' given as 'artifactsss'."""
        tpl_snippet = '''
        node_templates:
          mysql_database:
            type: tosca.nodes.Database
            artifactsss:
              db_content:
                implementation: files/my_db_content.txt
                type: tosca.artifacts.File
        '''
        expectedmessage = _('Node template "mysql_database" contains unknown '
                            'field "artifactsss". Refer to the definition to '
                            'verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_with_multiple_wrong_keynames(self):
        """Node templates given with multiple wrong keynames."""
        tpl_snippet = '''
        node_templates:
          mysql_dbms:
            type: tosca.nodes.DBMS
            propertieees:
              root_password: aaa
              port: 3376
            requirements:
              - host: server
            interfacs:
              Standard:
                configure: mysql_database_configure.sh
        '''
        expectedmessage = _('Node template "mysql_dbms" contains unknown '
                            'field "propertieees". Refer to the definition to '
                            'verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

        tpl_snippet = '''
        node_templates:
          mysql_database:
            type: tosca.nodes.Database
            properties:
              name: { get_input: db_name }
              user: { get_input: db_user }
              password: { get_input: db_pwd }
            capabilitiiiies:
              database_endpoint:
              properties:
                port: { get_input: db_port }
            requirementsss:
              - host:
                  node: mysql_dbms
            interfac:
              Standard:
                 configure: mysql_database_configure.sh

        '''
        expectedmessage = _('Node template "mysql_database" contains unknown '
                            'field "capabilitiiiies". Refer to the definition '
                            'to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_type(self):
        tpl_snippet = '''
        node_templates:
          mysql_database:
            type: tosca.nodes.Databases
            properties:
              db_name: { get_input: db_name }
              db_user: { get_input: db_user }
              db_password: { get_input: db_pwd }
            capabilities:
              database_endpoint:
                properties:
                  port: { get_input: db_port }
            requirements:
              - host: mysql_dbms
            interfaces:
              Standard:
                 configure: mysql_database_configure.sh
        '''
        expectedmessage = _('Type "tosca.nodes.Databases" is not '
                            'a valid type.')
        err = self.assertRaises(
            exception.InvalidTypeError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_requirements(self):
        tpl_snippet = '''
        node_templates:
          webserver:
            type: tosca.nodes.WebServer
            requirements:
              host: server
            interfaces:
              Standard:
                create: webserver_install.sh
                start: d.sh
        '''
        expectedmessage = _('"requirements" of template "webserver" must be '
                            'of type "list".')
        err = self.assertRaises(
            exception.TypeMismatchError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

        tpl_snippet = '''
        node_templates:
          mysql_database:
            type: tosca.nodes.Database
            properties:
              db_name: { get_input: db_name }
              db_user: { get_input: db_user }
              db_password: { get_input: db_pwd }
            capabilities:
              database_endpoint:
                properties:
                  port: { get_input: db_port }
            requirements:
              - host: mysql_dbms
              - database_endpoint: mysql_database
            interfaces:
              Standard:
                 configure: mysql_database_configure.sh
        '''
        expectedmessage = _('"requirements" of template "mysql_database" '
                            'contains unknown field "database_endpoint". '
                            'Refer to the definition to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_requirements_with_wrong_node_keyname(self):
        """Node template requirements keyname 'node' given as 'nodes'."""
        tpl_snippet = '''
        node_templates:
          mysql_database:
            type: tosca.nodes.Database
            requirements:
              - host:
                  nodes: mysql_dbms

        '''
        expectedmessage = _('"requirements" of template "mysql_database" '
                            'contains unknown field "nodes". Refer to the '
                            'definition to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_requirements_with_wrong_capability_keyname(self):
        """Incorrect node template requirements keyname

        Node template requirements keyname 'capability' given as
        'capabilityy'.
        """
        tpl_snippet = '''
        node_templates:
          mysql_database:
            type: tosca.nodes.Database
            requirements:
              - host:
                  node: mysql_dbms
              - log_endpoint:
                  node: logstash
                  capabilityy: log_endpoint
                  relationship:
                    type: tosca.relationships.ConnectsTo

        '''
        expectedmessage = _('"requirements" of template "mysql_database" '
                            'contains unknown field "capabilityy". Refer to '
                            'the definition to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_requirements_with_wrong_relationship_keyname(self):
        """Incorrect node template requirements keyname

        Node template requirements keyname 'relationship' given as
        'relationshipppp'.
        """
        tpl_snippet = '''
        node_templates:
          mysql_database:
            type: tosca.nodes.Database
            requirements:
              - host:
                  node: mysql_dbms
              - log_endpoint:
                  node: logstash
                  capability: log_endpoint
                  relationshipppp:
                    type: tosca.relationships.ConnectsTo

        '''
        expectedmessage = _('"requirements" of template "mysql_database" '
                            'contains unknown field "relationshipppp". Refer '
                            'to the definition to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_requirements_with_wrong_occurrences_keyname(self):
        """Incorrect node template requirements keyname

        Node template requirements keyname 'occurrences' given as
        'occurences'.
        """
        tpl_snippet = '''
        node_templates:
          mysql_database:
            type: tosca.nodes.Database
            requirements:
              - host:
                  node: mysql_dbms
              - log_endpoint:
                  node: logstash
                  capability: log_endpoint
                  relationship:
                    type: tosca.relationships.ConnectsTo
                  occurences: [0, UNBOUNDED]
        '''
        expectedmessage = _('"requirements" of template "mysql_database" '
                            'contains unknown field "occurences". Refer to '
                            'the definition to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_requirements_with_multiple_wrong_keynames(self):
        """Node templates given with multiple wrong requirements keynames."""
        tpl_snippet = '''
        node_templates:
          mysql_database:
            type: tosca.nodes.Database
            requirements:
              - host:
                  node: mysql_dbms
              - log_endpoint:
                  nod: logstash
                  capabilit: log_endpoint
                  relationshipppp:
                    type: tosca.relationships.ConnectsTo

        '''
        expectedmessage = _('"requirements" of template "mysql_database" '
                            'contains unknown field "nod". Refer to the '
                            'definition to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

        tpl_snippet = '''
        node_templates:
          mysql_database:
            type: tosca.nodes.Database
            requirements:
              - host:
                  node: mysql_dbms
              - log_endpoint:
                  node: logstash
                  capabilit: log_endpoint
                  relationshipppp:
                    type: tosca.relationships.ConnectsTo

        '''
        expectedmessage = _('"requirements" of template "mysql_database" '
                            'contains unknown field "capabilit". Refer to the '
                            'definition to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_requirements_invalid_occurrences(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            requirements:
              - log_endpoint:
                  capability: log_endpoint
                  occurrences: [0, -1]
        '''
        expectedmessage = _('Value of property "[0, -1]" is invalid.')
        err = self.assertRaises(
            exception.InvalidPropertyValueError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            requirements:
              - log_endpoint:
                  capability: log_endpoint
                  occurrences: [a, w]
        '''
        expectedmessage = _('"a" is not an integer.')
        err = self.assertRaises(
            ValueError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            requirements:
              - log_endpoint:
                  capability: log_endpoint
                  occurrences: -1
        '''
        expectedmessage = _('"-1" is not a list.')
        err = self.assertRaises(
            ValueError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            requirements:
              - log_endpoint:
                  capability: log_endpoint
                  occurrences: [5, 1]
        '''
        expectedmessage = _('Value of property "[5, 1]" is invalid.')
        err = self.assertRaises(
            exception.InvalidPropertyValueError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            requirements:
              - log_endpoint:
                  capability: log_endpoint
                  occurrences: [0, 0]
        '''
        expectedmessage = _('Value of property "[0, 0]" is invalid.')
        err = self.assertRaises(
            exception.InvalidPropertyValueError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_requirements_valid_occurrences(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            requirements:
              - log_endpoint:
                  capability: log_endpoint
                  occurrences: [2, 2]
        '''
        self._single_node_template_content_test(tpl_snippet)

    def test_node_template_capabilities(self):
        tpl_snippet = '''
        node_templates:
          mysql_database:
            type: tosca.nodes.Database
            properties:
              db_name: { get_input: db_name }
              db_user: { get_input: db_user }
              db_password: { get_input: db_pwd }
            capabilities:
              http_endpoint:
                properties:
                  port: { get_input: db_port }
            requirements:
              - host: mysql_dbms
            interfaces:
              Standard:
                 configure: mysql_database_configure.sh
        '''
        expectedmessage = _('"capabilities" of template "mysql_database" '
                            'contains unknown field "http_endpoint". Refer to '
                            'the definition to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_properties(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            properties:
              os_image: F18_x86_64
            capabilities:
              host:
                properties:
                  disk_size: 10 GB
                  num_cpus: { get_input: cpus }
                  mem_size: 4096 MB
              os:
                properties:
                  architecture: x86_64
                  type: Linux
                  distribution: Fedora
                  version: 18.0
        '''
        expectedmessage = _('"properties" of template "server" contains '
                            'unknown field "os_image". Refer to the '
                            'definition to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_interfaces(self):
        tpl_snippet = '''
        node_templates:
          wordpress:
            type: tosca.nodes.WebApplication.WordPress
            requirements:
              - host: webserver
              - database_endpoint: mysql_database
            interfaces:
              Standards:
                 create: wordpress_install.sh
                 configure:
                   implementation: wordpress_configure.sh
                   inputs:
                     wp_db_name: { get_property: [ mysql_database, db_name ] }
                     wp_db_user: { get_property: [ mysql_database, db_user ] }
                     wp_db_password: { get_property: [ mysql_database, \
                     db_password ] }
                     wp_db_port: { get_property: [ SELF, \
                     database_endpoint, port ] }
        '''
        expectedmessage = _('"interfaces" of template "wordpress" contains '
                            'unknown field "Standards". Refer to the '
                            'definition to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

        tpl_snippet = '''
        node_templates:
          wordpress:
            type: tosca.nodes.WebApplication.WordPress
            requirements:
              - host: webserver
              - database_endpoint: mysql_database
            interfaces:
              Standard:
                 create: wordpress_install.sh
                 config:
                   implementation: wordpress_configure.sh
                   inputs:
                     wp_db_name: { get_property: [ mysql_database, db_name ] }
                     wp_db_user: { get_property: [ mysql_database, db_user ] }
                     wp_db_password: { get_property: [ mysql_database, \
                     db_password ] }
                     wp_db_port: { get_property: [ SELF, \
                     database_endpoint, port ] }
        '''
        expectedmessage = _('"interfaces" of template "wordpress" contains '
                            'unknown field "config". Refer to the definition '
                            'to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

        tpl_snippet = '''
        node_templates:
          wordpress:
            type: tosca.nodes.WebApplication.WordPress
            requirements:
              - host: webserver
              - database_endpoint: mysql_database
            interfaces:
              Standard:
                 create: wordpress_install.sh
                 configure:
                   implementation: wordpress_configure.sh
                   input:
                     wp_db_name: { get_property: [ mysql_database, db_name ] }
                     wp_db_user: { get_property: [ mysql_database, db_user ] }
                     wp_db_password: { get_property: [ mysql_database, \
                     db_password ] }
                     wp_db_port: { get_ref_property: [ database_endpoint, \
                     database_endpoint, port ] }
        '''
        expectedmessage = _('"interfaces" of template "wordpress" contains '
                            'unknown field "input". Refer to the definition '
                            'to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_relationship_template_properties(self):
        tpl_snippet = '''
        relationship_templates:
            storage_attachto:
                type: AttachesTo
                properties:
                  device: test_device
        '''
        expectedmessage = _('"properties" of template "storage_attachto" is '
                            'missing required field "[\'location\']".')
        rel_template = (toscaparser.utils.yamlparser.
                        simple_parse(tpl_snippet))['relationship_templates']
        name = list(rel_template.keys())[0]
        rel_template = RelationshipTemplate(rel_template[name], name)
        err = self.assertRaises(exception.MissingRequiredFieldError,
                                rel_template.validate)
        self.assertEqual(expectedmessage, str(err))

    def test_relationship_template_properties_key_schema(self):
        tpl_path = utils.get_sample_test_path(
            "data/relationship/test_custom_relationship_key_schema.yaml")
        tosca = ToscaTemplate(tpl_path)
        p_key_schema_type = tosca.tpl['relationship_types'][
            'tosca.nodes.CustomRelationshipTest']['properties'][
                'property_test']['key_schema']['type']
        self.assertEqual('string', p_key_schema_type)
        p_key_schema_type = tosca.relationship_templates[0].custom_def[
            'tosca.nodes.CustomRelationshipTest']['properties'][
                'property_test']['key_schema']['type']
        self.assertEqual('string', p_key_schema_type)
        property_value = (
            tosca.topology_template.relationship_templates[0].entity_tpl[
                'properties']['property_test']['test_key'])
        self.assertEqual('test_value', property_value)

    def test_tosca_version_1_3(self):
        tpl_path = utils.get_sample_test_path(
            "data/test_tosca_version_1_3.yaml")
        self.assertIsNotNone(ToscaTemplate(tpl_path))

    def test_import_tosca_version_1_3(self):
        tpl_path = utils.get_sample_test_path(
            "data/test_import_tosca_version_1_3.yaml")
        self.assertIsNotNone(ToscaTemplate(tpl_path))

    def test_invalid_template_version(self):
        tosca_tpl = utils.get_sample_test_path(
            "data/test_invalid_template_version.yaml")
        self.assertRaises(exception.ValidationError, ToscaTemplate, tosca_tpl)
        valid_versions = '", "'.join(ToscaTemplate.VALID_TEMPLATE_VERSIONS)
        exception.ExceptionCollector.assertExceptionMessage(
            exception.InvalidTemplateVersion,
            (_('The template version "tosca_xyz" is invalid. Valid versions '
               'are "%s".') % valid_versions))

    def test_import_invalid_template_version(self):
        tosca_tpl = utils.get_sample_test_path(
            "data/test_import_invalid_template_version.yaml")
        self.assertRaises(exception.ValidationError, ToscaTemplate, tosca_tpl)
        valid_versions = '", "'.join(ToscaTemplate.VALID_TEMPLATE_VERSIONS)
        exception.ExceptionCollector.assertExceptionMessage(
            exception.InvalidTemplateVersion,
            (_('The template version "tosca_simple_yaml_XXX in '
               '{\'invalid\': \'custom_types/invalid_template_version.yaml\'}"'
               ' is invalid. Valid versions are "%s".') % valid_versions))

    def test_import_template_metadata(self):
        tosca_tpl = utils.get_sample_test_path("data/test_import_metadata.yml")
        ToscaTemplate(tosca_tpl)

    def test_node_template_capabilities_properties(self):
        # validating capability property values
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.WebServer
            capabilities:
              data_endpoint:
                properties:
                  initiator: test
        '''
        expectedmessage = _('The value "test" of property "initiator" is '
                            'not valid. Expected a value from "[source, '
                            'target, peer]".')

        err = self.assertRaises(
            exception.ValidationError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            capabilities:
              host:
                properties:
                  disk_size: 10 GB
                  num_cpus: { get_input: cpus }
                  mem_size: 4096 MB
              os:
                properties:
                  architecture: x86_64
                  type: Linux
                  distribution: Fedora
                  version: 18.0
              scalable:
                properties:
                  min_instances: 1
                  max_instances: 3
                  default_instances: 5
        '''
        expectedmessage = _('"properties" of template "server": '
                            '"default_instances" value is not between '
                            '"min_instances" and "max_instances".')
        err = self.assertRaises(
            exception.ValidationError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_objectstorage_without_required_property(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.ObjectStorage
            properties:
              maxsize: 1 GB
        '''
        expectedmessage = _('"properties" of template "server" is missing '
                            'required field "[\'name\']".')
        err = self.assertRaises(
            exception.MissingRequiredFieldError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_objectstorage_with_invalid_scalar_unit(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.ObjectStorage
            properties:
              name: test
              maxsize: -1
        '''
        expectedmessage = _('"-1" is not a valid scalar-unit.')
        err = self.assertRaises(
            ValueError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_node_template_objectstorage_with_invalid_scalar_type(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.ObjectStorage
            properties:
              name: test
              maxsize: 1 XB
        '''
        expectedmessage = _('"1 XB" is not a valid scalar-unit.')
        err = self.assertRaises(
            ValueError,
            lambda: self._single_node_template_content_test(tpl_snippet))
        self.assertEqual(expectedmessage, err.__str__())

    def test_special_keywords(self):
        """Test special keywords

           Test that special keywords, e.g. metadata, which are not part
           of specification do not throw any validation error.
        """
        tpl_snippet_metadata_map = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            metadata:
              name: server A
              role: master
        '''
        self._single_node_template_content_test(tpl_snippet_metadata_map)

        tpl_snippet_metadata_inline = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            metadata: none
        '''
        self._single_node_template_content_test(tpl_snippet_metadata_inline)

    def test_policy_valid_keynames(self):
        tpl_snippet = '''
        policies:
          - servers_placement:
              type: tosca.policies.Placement
              description: Apply placement policy to servers
              metadata: { user1: 1001, user2: 1002 }
              targets: [ serv1, serv2 ]
        '''
        policies = (toscaparser.utils.yamlparser.
                    simple_parse(tpl_snippet))['policies'][0]
        name = list(policies.keys())[0]
        Policy(name, policies[name], None, None)

    def test_policy_invalid_keyname(self):
        tpl_snippet = '''
        policies:
          - servers_placement:
              type: tosca.policies.Placement
              testkey: testvalue
        '''
        policies = (toscaparser.utils.yamlparser.
                    simple_parse(tpl_snippet))['policies'][0]
        name = list(policies.keys())[0]

        expectedmessage = _('Policy "servers_placement" contains '
                            'unknown field "testkey". Refer to the '
                            'definition to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: Policy(name, policies[name], None, None))
        self.assertEqual(expectedmessage, err.__str__())

    def test_policy_trigger_valid_keyname_senlin_resources(self):
        tpl_snippet = '''
        triggers:
         - resize_compute:
             description: trigger
             event_type: tosca.events.resource.utilization
             schedule:
               start_time: "2015-05-07T07:00:00Z"
               end_time: "2015-06-07T07:00:00Z"
             target_filter:
               node: master-container
               requirement: host
               capability: Container
             condition:
               constraint: { greater_than: 50 }
               granularity: 60
               evaluations: 1
               aggregation_method : mean
             action:
               resize: # Operation name
                inputs:
                 strategy: LEAST_USED
                 implementation: Senlin.webhook()
        '''
        triggers = (toscaparser.utils.yamlparser.
                    simple_parse(tpl_snippet))['triggers'][0]
        name = list(triggers.keys())[0]
        Triggers(name, triggers[name])

    # Since the keyname 'event_type' has been changed in TOSCA1.3,
    # test whether the new keyname 'event' can be used.
    def test_policy_trigger_keyname_event(self):
        tpl_snippet = '''
        triggers:
         - resize_compute:
             description: trigger
             event: tosca.events.resource.utilization
             schedule:
               start_time: "2015-05-07T07:00:00Z"
               end_time: "2015-06-07T07:00:00Z"
             target_filter:
               node: master-container
               requirement: host
               capability: Container
             condition:
               constraint: { greater_than: 50 }
               granularity: 60
               evaluations: 1
               aggregation_method : mean
             action:
               resize: # Operation name
                inputs:
                 strategy: LEAST_USED
                 implementation: Senlin.webhook()
        '''
        triggers = (toscaparser.utils.yamlparser.
                    simple_parse(tpl_snippet))['triggers'][0]
        name = list(triggers.keys())[0]
        trigger = Triggers(name, triggers[name])
        self.assertEqual('tosca.events.resource.utilization',
                         trigger.get_event())

    def test_policy_trigger_valid_keyname_heat_resources(self):
        tpl_snippet = '''
        triggers:
         - high_cpu_usage:
             description: trigger
             metric: cpu_util
             condition:
               constraint: utilization greater_than 60%
               threshold: 60
               granularity: 600
               evaluations: 1
               aggregation_method: mean
               resource_type: instance
               comparison_operator: gt
             metadata: SG1
             action: [SP1]
        '''
        triggers = (toscaparser.utils.yamlparser.
                    simple_parse(tpl_snippet))['triggers'][0]
        name = list(triggers.keys())[0]
        Triggers(name, triggers[name])

    def test_policy_trigger_invalid_keyname_senlin_resources(self):
        tpl_snippet = '''
        triggers:
         - resize_compute:
             description: trigger
             event_type: tosca.events.resource.utilization
             schedule:
               start_time: "2015-05-07T07:00:00Z"
               end_time: "2015-06-07T07:00:00Z"
             target_filter1:
               node: master-container
               requirement: host
               capability: Container
             condition:
               constraint: utilization greater_than 50%
               granularity1: 60
               evaluations: 1
               aggregation_method: mean
               resource_type: instance
             action:
               resize: # Operation name
                inputs:
                 strategy: LEAST_USED
                 implementation: Senlin.webhook()
        '''
        triggers = (toscaparser.utils.yamlparser.
                    simple_parse(tpl_snippet))['triggers'][0]
        name = list(triggers.keys())[0]
        expectedmessage = _(
            'Triggers "resize_compute" contains unknown field '
            '"target_filter1". Refer to the definition '
            'to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: Triggers(name, triggers[name]))
        self.assertEqual(expectedmessage, err.__str__())

    def test_policy_trigger_invalid_keyname_heat_resources(self):
        tpl_snippet = '''
        triggers:
         - high_cpu_usage:
             description: trigger
             metric: cpu_util
             condition:
               constraint: utilization greater_than 60%
               threshold: 60
               granularity: 600
               evaluations: 1
               aggregation_method: mean
               resource_type: instance
               comparison_operator: gt
             metadata1: SG1
             action: [SP1]
        '''
        triggers = (toscaparser.utils.yamlparser.
                    simple_parse(tpl_snippet))['triggers'][0]
        name = list(triggers.keys())[0]
        expectedmessage = _(
            'Triggers "high_cpu_usage" contains unknown field '
            '"metadata1". Refer to the definition '
            'to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: Triggers(name, triggers[name]))
        self.assertEqual(expectedmessage, err.__str__())

    def test_policy_missing_required_keyname(self):
        tpl_snippet = '''
        policies:
          - servers_placement:
              description: test description
        '''
        policies = (toscaparser.utils.yamlparser.
                    simple_parse(tpl_snippet))['policies'][0]
        name = list(policies.keys())[0]

        expectedmessage = _('Template "servers_placement" is missing '
                            'required field "type".')
        err = self.assertRaises(
            exception.MissingRequiredFieldError,
            lambda: Policy(name, policies[name], None, None))
        self.assertEqual(expectedmessage, err.__str__())

    def test_policy_without_required_property(self):
        tpl_snippet = '''
        policies:
          - some_policy:
              type: tosca.policies.somePolicy
              properties:
                value: 100
        '''
        policies = (toscaparser.utils.yamlparser.
                    simple_parse(tpl_snippet))['policies'][0]
        name = list(policies.keys())[0]
        policyObj = Policy(name, policies[name], None, None,
                           self._custom_types_policy())
        expectedmessage = _('"properties" of template "some_policy" is '
                            'missing required field "[\'name\']".')
        err = self.assertRaises(
            exception.MissingRequiredFieldError,
            policyObj.validate)
        self.assertEqual(expectedmessage, err.__str__())

    def test_credential_datatype(self):
        tosca_tpl = utils.get_sample_test_path(
            "data/test_credential_datatype.yaml")
        self.assertIsNotNone(ToscaTemplate(tosca_tpl))

    def test_invalid_default_value(self):
        tpl_path = utils.get_sample_test_path(
            "data/test_invalid_input_defaults.yaml")
        self.assertRaises(exception.ValidationError, ToscaTemplate, tpl_path)
        exception.ExceptionCollector.assertExceptionMessage(
            ValueError, _('"two" is not an integer.'))

    def test_invalid_capability(self):
        tpl_snippet = '''
        node_templates:
          server:
            type: tosca.nodes.Compute
            capabilities:
                oss:
                    properties:
                        architecture: x86_64
        '''
        tpl = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        err = self.assertRaises(exception.UnknownFieldError,
                                TopologyTemplate, tpl, None)
        expectedmessage = _('"capabilities" of template "server" contains '
                            'unknown field "oss". Refer to the definition '
                            'to verify valid values.')
        self.assertEqual(expectedmessage, err.__str__())

    def test_qualified_name(self):
        tpl_snippet_full_name = '''
        node_templates:
          supported_type:
            type: tosca.nodes.Compute
        '''
        tpl = (
            toscaparser.utils.yamlparser.simple_parse(
                tpl_snippet_full_name))
        TopologyTemplate(tpl, None)

        tpl_snippet_short_name = '''
        node_templates:
          supported_type:
            type: Compute
        '''
        tpl = (
            toscaparser.utils.yamlparser.simple_parse(
                tpl_snippet_short_name))
        TopologyTemplate(tpl, None)

        tpl_snippet_qualified_name = '''
        node_templates:
          supported_type:
            type: tosca:Compute
        '''
        tpl = (
            toscaparser.utils.yamlparser.simple_parse(
                tpl_snippet_qualified_name))
        TopologyTemplate(tpl, None)

    def test_requirements_as_list(self):
        """Node template with requirements provided with or without list

        Node template requirements are required to be provided as list.
        """

        expectedmessage = _('"requirements" of template "my_webserver"'
                            ' must be of type "list".')

        # requirements provided as dictionary
        tpl_snippet1 = '''
        node_templates:
          my_webserver:
            type: tosca.nodes.WebServer
            requirements:
              host: server
          server:
            type: tosca.nodes.Compute
        '''
        err1 = self.assertRaises(
            exception.TypeMismatchError,
            lambda: self._single_node_template_content_test(tpl_snippet1))
        self.assertEqual(expectedmessage, err1.__str__())

        # requirements provided as string
        tpl_snippet2 = '''
        node_templates:
          my_webserver:
            type: tosca.nodes.WebServer
            requirements: server
          server:
            type: tosca.nodes.Compute
        '''
        err2 = self.assertRaises(
            exception.TypeMismatchError,
            lambda: self._single_node_template_content_test(tpl_snippet2))
        self.assertEqual(expectedmessage, err2.__str__())

        # requirements provided as list
        tpl_snippet3 = '''
        node_templates:
          my_webserver:
            type: tosca.nodes.WebServer
            requirements:
              - host: server
          server:
            type: tosca.nodes.Compute
        '''
        self.assertIsNone(
            self._single_node_template_content_test(tpl_snippet3))

    def test_properties_override_with_flavor_and_image(self):
        tpl_path = utils.get_sample_test_path(
            "data/test_normative_type_properties_override.yaml")
        self.assertIsNotNone(ToscaTemplate(tpl_path))

    def test_long_rel(self):
        tpl_path = utils.get_sample_test_path("data/test_long_rel.yaml")
        self.assertIsNotNone(ToscaTemplate(tpl_path))

    def test_policy_reservation_valid_keyname_heat_resources(self):
        tpl_snippet = '''
        reservation:
          start_actions: [SP_RSV]
          before_end_actions: [SP_RSV]
          end_actions: [noop]
          properties:
            lease_id: '58d2239c-f181-4915-bdcb-040f7ef911a7'
        '''
        reservation = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        name = list(reservation.keys())[0]
        Reservation(reservation[name])

    def test_policy_reservation_invalid_keyname_heat_resources(self):
        tpl_snippet = '''
        reservation:
          start_actions: [SP_RSV]
          before_actions: [SP_RSV]
          end_actions: [noop]
          properties:
            lease_id: '58d2239c-f181-4915-bdcb-040f7ef911a7'
        '''
        reservation = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        name = list(reservation.keys())[0]
        expectedmessage = _(
            'Reservation contains unknown field '
            '"before_actions". Refer to the definition '
            'to verify valid values.')
        err = self.assertRaises(
            exception.UnknownFieldError,
            lambda: Reservation(reservation[name]))
        self.assertEqual(expectedmessage, err.__str__())

    def test_policy_reservation_missing_key_heat_resources(self):
        tpl_snippet = '''
        reservation:
          start_actions: [SP_RSV]
          end_actions: [noop]
          properties:
            lease_id: '58d2239c-f181-4915-bdcb-040f7ef911a7'
        '''
        reservation = (toscaparser.utils.yamlparser.simple_parse(tpl_snippet))
        name = list(reservation.keys())[0]
        expectedmessage = _('Reservation is missing '
                            'required field "before_end_actions".')
        err = self.assertRaises(
            exception.MissingRequiredFieldError,
            lambda: Reservation(reservation[name]))
        self.assertEqual(expectedmessage, err.__str__())

    def test_scalar_unit_without_unit(self):
        tpl_path = utils.get_sample_test_path(
            "data/test_scalar_unit_without_unit.yaml")
        self.assertRaises(exception.ValidationError,
                          lambda: ToscaTemplate(tpl_path))

    def test_template_without_requirement(self):
        tpl_path = utils.get_sample_test_path(
            "data/test_template_without_requirement.yaml")
        err = self.assertRaises(exception.ValidationError,
                                lambda: ToscaTemplate(tpl_path))
        expectedmessage = _('Relationship "host" in template "webserver" has '
                            'a wrong number of occurrences')
        self.assertIn(expectedmessage, err.message)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from toscaparser.common.exception import ExceptionCollector
from toscaparser.common.exception import UnknownFieldError
from toscaparser.entity_template import EntityTemplate
from toscaparser.utils import validateutils

# In TOSCA1.3, the section name 'interfaces' is not defined for groups.
# It will be ignored if the 'interfaces' is defined as TOSCA1.3.
SECTIONS = (TYPE, METADATA, DESCRIPTION, PROPERTIES, MEMBERS, INTERFACES,
            ATTRIBUTES) = \
           ('type', 'metadata', 'description',
            'properties', 'members', 'interfaces', 'attributes')


class Group(EntityTemplate):

    def __init__(self, name, group_templates, member_nodes, custom_defs=None):
        super(Group, self).__init__(name,
                                    group_templates,
                                    'group_type',
                                    custom_defs)
        self.name = name
        self.tpl = group_templates
        self.meta_data = None
        if self.METADATA in self.tpl:
            self.meta_data = self.tpl.get(self.METADATA)
            validateutils.validate_map(self.meta_data)
        self.member_nodes = member_nodes
        self._validate_keys()

    @property
    def members(self):
        return self.entity_tpl.get('members')

    @property
    def description(self):
        return self.entity_tpl.get('description')

    def get_member_nodes(self):
        return self.member_nodes

    def _validate_keys(self):
        for key in self.entity_tpl.keys():
            if key not in SECTIONS:
                ExceptionCollector.appendException(
                    UnknownFieldError(what='Groups "%s"' % self.name,
                                      field=key))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from testtools import matchers

from toscaparser.common import exception
from toscaparser.elements.property_definition import PropertyDef
from toscaparser.nodetemplate import NodeTemplate
from toscaparser.properties import Property
from toscaparser.tests.base import TestCase
from toscaparser.utils.gettextutils import _
from toscaparser.utils import yamlparser


class PropertyTest(TestCase):

    def test_type(self):
        test_property_schema = {'type': 'string'}
        propertyInstance = Property('test_property', 'Hughes',
                                    test_property_schema)
        self.assertEqual('string', propertyInstance.type)

    def test_type_invalid(self):
        test_property_schema = {'type': 'Fish'}
        propertyInstance = Property('test_property', 'Hughes',
                                    test_property_schema)
        error = self.assertRaises(exception.InvalidTypeError,
                                  propertyInstance.validate)
        self.assertEqual(_('Type "Fish" is not a valid type.'), str(error))

    def test_list(self):
        test_property_schema = {'type': 'list'}
        propertyInstance = Property('test_property', ['a', 'b'],
                                    test_property_schema)
        self.assertIsNone(propertyInstance.validate())
        self.assertEqual(['a', 'b'], propertyInstance.value)

    def test_list_invalid(self):
        test_property_schema = {'type': 'list'}
        propertyInstance = Property('test_property', 'a',
                                    test_property_schema)
        error = self.assertRaises(ValueError, propertyInstance.validate)
        self.assertEqual(_('"a" is not a list.'), str(error))

    def test_list_entry_schema(self):
        test_property_schema = {'type': 'list',
                                'entry_schema': {'type': 'string'}}
        propertyInstance = Property('test_property', ['a', 'b'],
                                    test_property_schema)
        self.assertIsNone(propertyInstance.validate())
        self.assertEqual(['a', 'b'], propertyInstance.value)

        schema_snippet = '''
        type: list
        entry_schema:
          type: string
          constraints:
            - min_length: 2
        '''
        test_property_schema = yamlparser.simple_parse(schema_snippet)
        propertyInstance = Property('test_property', ['ab', 'cd'],
                                    test_property_schema)
        self.assertIsNone(propertyInstance.validate())
        self.assertEqual(['ab', 'cd'], propertyInstance.value)

    def test_list_entry_schema_invalid(self):
        test_property_schema = {'type': 'list',
                                'entry_schema': {'type': 'integer'}}
        propertyInstance = Property('test_property', [1, 'b'],
                                    test_property_schema)
        error = self.assertRaises(ValueError, propertyInstance.validate)
        self.assertEqual(_('"b" is not an integer.'), str(error))

    def test_map(self):
        test_property_schema = {'type': 'map'}
        propertyInstance = Property('test_property', {'a': 'b'},
                                    test_property_schema)
        self.assertIsNone(propertyInstance.validate())
        self.assertEqual({'a': 'b'}, propertyInstance.value)

    def test_map_invalid(self):
        test_property_schema = {'type': 'map'}
        propertyInstance = Property('test_property', 12,
                                    test_property_schema)
        error = self.assertRaises(ValueError, propertyInstance.validate)
        self.assertEqual(_('"12" is not a map.'), str(error))

    def test_map_entry_schema(self):
        test_property_schema = {'type': 'map',
                                'entry_schema': {'type': 'boolean'}}
        propertyInstance = Property('test_property',
                                    {'valid': True, 'required': True},
                                    test_property_schema)
        self.assertIsNone(propertyInstance.validate())
        self.assertEqual({'valid': True, 'required': True},
                         propertyInstance.value)

    def test_map_entry_schema_invalid(self):
        test_property_schema = {'type': 'map',
                                'entry_schema': {'type': 'boolean'}}
        propertyInstance = Property('test_property',
                                    {'valid': True, 'contact_name': 123},
                                    test_property_schema)
        error = self.assertRaises(ValueError, propertyInstance.validate)
        self.assertEqual(_('"123" is not a boolean.'), str(error))

    def test_boolean(self):
        test_property_schema = {'type': 'boolean'}
        propertyInstance = Property('test_property', 'true',
                                    test_property_schema)
        self.assertIsNone(propertyInstance.validate())
        propertyInstance = Property('test_property', True,
                                    test_property_schema)
        self.assertIsNone(propertyInstance.validate())
        self.assertEqual(True, propertyInstance.value)

    def test_boolean_invalid(self):
        test_property_schema = {'type': 'boolean'}
        propertyInstance = Property('test_property', 12,
                                    test_property_schema)
        error = self.assertRaises(ValueError, propertyInstance.validate)
        self.assertEqual(_('"12" is not a boolean.'), str(error))

    def test_float(self):
        test_property_schema = {'type': 'float'}
        propertyInstance = Property('test_property', 0.1,
                                    test_property_schema)
        self.assertIsNone(propertyInstance.validate())
        self.assertEqual(0.1, propertyInstance.value)

    def test_float_invalid(self):
        test_property_schema = {'type': 'float'}
        propertyInstance = Property('test_property', 12,
                                    test_property_schema)
        error = self.assertRaises(ValueError, propertyInstance.validate)
        self.assertEqual(_('"12" is not a float.'), str(error))

    def test_timestamp(self):
        test_property_schema = {'type': 'timestamp'}
        # canonical timestamp
        propertyInstance = Property('test_property', '2015-04-01T02:59:43.1Z',
                                    test_property_schema)
        self.assertIsNone(propertyInstance.validate())
        self.assertEqual("2015-04-01T02:59:43.1Z", propertyInstance.value)

        # iso8601 timestamp
        propertyInstance = Property('test_property',
                                    '2015-04-01t21:59:43.10-05:00',
                                    test_property_schema)
        self.assertIsNone(propertyInstance.validate())
        self.assertEqual("2015-04-01t21:59:43.10-05:00",
                         propertyInstance.value)

        # space separated timestamp
        propertyInstance = Property('test_property',
                                    '2015-04-01 21:59:43.10 -5',
                                    test_property_schema)
        self.assertIsNone(propertyInstance.validate())
        self.assertEqual("2015-04-01 21:59:43.10 -5", propertyInstance.value)

        # no time zone timestamp
        propertyInstance = Property('test_property', '2015-04-01 21:59:43.10',
                                    test_property_schema)
        self.assertIsNone(propertyInstance.validate())
        self.assertEqual("2015-04-01 21:59:43.10", propertyInstance.value)

        # date (00:00:00Z)
        propertyInstance = Property('test_property', '2015-04-01',
                                    test_property_schema)
        self.assertIsNone(propertyInstance.validate())
        self.assertEqual("2015-04-01", propertyInstance.value)

    def test_timestamp_invalid(self):
        test_property_schema = {'type': 'timestamp'}
        # invalid timestamp - day out of range
        value = '2015-04-115T02:59:43.1Z'
        propertyInstance = Property('test_property', value,
                                    test_property_schema)
        error = self.assertRaises(ValueError, propertyInstance.validate)
        expected_message = (_('"%s" is not a valid timestamp.') % value)
        self.assertThat(str(error), matchers.StartsWith(expected_message))

    def test_required(self):
        test_property_schema = {'type': 'string'}
        propertyInstance = Property('test_property', 'Foo',
                                    test_property_schema)
        self.assertEqual(True, propertyInstance.required)

    def test_proprety_inheritance(self):

        tosca_custom_def = '''
          tosca.nodes.SoftwareComponent.MySoftware:
            derived_from: SoftwareComponent
            properties:
              install_path:
                required: false
                type: string
                default: /opt/mysoftware
        '''

        tosca_node_template = '''
          node_templates:
            mysoftware_instance:
              type: tosca.nodes.SoftwareComponent.MySoftware
              properties:
                component_version: 3.1
              requirements:
                - host: my_server_host

            my_server_host:
              type: tosca.nodes.Compute
        '''

        expected_properties = ['component_version',
                               'install_path']

        tpl = self._get_nodetemplate(tosca_node_template, tosca_custom_def)
        self.assertIsNone(tpl.validate())
        self.assertEqual(expected_properties,
                         sorted(tpl.get_properties().keys()))

    def test_missing_property_type(self):
        tpl_snippet = '''
         properties:
           prop:
             typo: tosca.mytesttype.Test
        '''
        schema = yamlparser.simple_parse(tpl_snippet)
        error = self.assertRaises(exception.InvalidSchemaError, PropertyDef,
                                  'prop', None, schema['properties']['prop'])
        self.assertEqual(_('Schema definition of "prop" must have a "type" '
                           'attribute.'), str(error))

    def test_invalid_required_value(self):
        tpl_snippet = '''
         properties:
           prop:
             type: tosca.mytesttype.Test
             required: dunno
        '''
        schema = yamlparser.simple_parse(tpl_snippet)
        error = self.assertRaises(exception.InvalidSchemaError, PropertyDef,
                                  'prop', None, schema['properties']['prop'])

        valid_values = ', '.join(PropertyDef.VALID_REQUIRED_VALUES)
        expected_message = (_('Schema definition of "prop" has "required" '
                              'attribute with invalid value "dunno". The '
                              'value must be one of "%s".') % valid_values)
        self.assertEqual(expected_message, str(error))

    def test_invalid_property_status(self):
        tpl_snippet = '''
         properties:
           prop:
             type: string
             status: unknown
        '''
        schema = yamlparser.simple_parse(tpl_snippet)
        error = self.assertRaises(exception.InvalidSchemaError, PropertyDef,
                                  'prop', None, schema['properties']['prop'])

        valid_values = ', '.join(PropertyDef.VALID_STATUS_VALUES)
        expected_message = (_('Schema definition of "prop" has "status" '
                              'attribute with invalid value "unknown". The '
                              'value must be one of "%s".') % valid_values)
        self.assertEqual(expected_message, str(error))

    def test_capability_proprety_inheritance(self):
        tosca_custom_def_example1 = '''
          tosca.capabilities.ScalableNew:
            derived_from: tosca.capabilities.Scalable
            properties:
              max_instances:
                type: integer
                default: 0
                required: no

          tosca.nodes.ComputeNew:
            derived_from: tosca.nodes.Compute
            capabilities:
              scalable:
                type: tosca.capabilities.ScalableNew
        '''

        tosca_node_template_example1 = '''
          node_templates:
            compute_instance:
              type: tosca.nodes.ComputeNew
              capabilities:
                scalable:
                  properties:
                    min_instances: 1
        '''

        tosca_custom_def_example2 = '''
          tosca.nodes.ComputeNew:
            derived_from: tosca.nodes.Compute
            capabilities:
              new_cap:
                type: tosca.capabilities.Scalable
        '''

        tosca_node_template_example2 = '''
          node_templates:
            db_server:
                type: tosca.nodes.ComputeNew
                capabilities:
                  host:
                   properties:
                     num_cpus: 1
        '''

        tpl1 = self._get_nodetemplate(tosca_node_template_example1,
                                      tosca_custom_def_example1)
        self.assertIsNone(tpl1.validate())

        tpl2 = self._get_nodetemplate(tosca_node_template_example2,
                                      tosca_custom_def_example2)
        self.assertIsNone(tpl2.validate())

    def _get_nodetemplate(self, tpl_snippet,
                          custom_def_snippet=None):
        nodetemplates = yamlparser.\
            simple_parse(tpl_snippet)['node_templates']
        custom_def = []
        if custom_def_snippet:
            custom_def = yamlparser.simple_parse(custom_def_snippet)
        name = list(nodetemplates.keys())[0]
        tpl = NodeTemplate(name, nodetemplates, custom_def)
        return tpl

    def test_explicit_relationship_proprety(self):

        tosca_node_template = '''
          node_templates:

            client_node:
              type: tosca.nodes.Compute
              requirements:
                - local_storage:
                    node: my_storage
                    relationship:
                      type: AttachesTo
                      properties:
                        location: /mnt/disk

            my_storage:
              type: tosca.nodes.BlockStorage
              properties:
                size: 1 GB
        '''

        expected_properties = ['location']

        nodetemplates = yamlparser.\
            simple_parse(tosca_node_template)['node_templates']
        tpl = NodeTemplate('client_node', nodetemplates, [])

        self.assertIsNone(tpl.validate())
        rel_tpls = []
        for relationship, trgt in tpl.relationships.items():
            rel_tpls.extend(trgt.get_relationship_template())
        self.assertEqual(expected_properties,
                         sorted(rel_tpls[0].get_properties().keys()))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from toscaparser.elements.property_definition import PropertyDef
from toscaparser.elements.statefulentitytype import StatefulEntityType


class CapabilityTypeDef(StatefulEntityType):
    '''TOSCA built-in capabilities type.'''
    TOSCA_TYPEURI_CAPABILITY_ROOT = 'tosca.capabilities.Root'

    def __init__(self, name, ctype, ntype, custom_def=None):
        self.name = name
        super(CapabilityTypeDef, self).__init__(ctype, self.CAPABILITY_PREFIX,
                                                custom_def)
        self.nodetype = ntype
        self.properties = None
        self.custom_def = custom_def
        if self.PROPERTIES in self.defs:
            self.properties = self.defs[self.PROPERTIES]
        self.parent_capabilities = self._get_parent_capabilities(custom_def)

    def get_properties_def_objects(self):
        '''Return a list of property definition objects.'''
        properties = []
        parent_properties = {}
        if self.parent_capabilities:
            for type, value in self.parent_capabilities.items():
                if self.PROPERTIES in value:
                    parent_properties[type] = value.get(self.PROPERTIES)
        if self.properties:
            for prop, schema in self.properties.items():
                properties.append(PropertyDef(prop, None, schema))
        if parent_properties:
            for parent, props in parent_properties.items():
                for prop, schema in props.items():
                    # add parent property if not overridden by children type
                    if not self.properties or \
                            prop not in self.properties.keys():
                        properties.append(PropertyDef(prop, None, schema))
        return properties

    def get_properties_def(self):
        '''Return a dictionary of property definition name-object pairs.'''
        return {prop.name: prop
                for prop in self.get_properties_def_objects()}

    def get_property_def_value(self, name):
        '''Return the definition of a given property name.'''
        props_def = self.get_properties_def()
        if props_def and name in props_def:
            return props_def[name].value

    def _get_parent_capabilities(self, custom_def=None):
        capabilities = {}
        parent_cap = self.parent_type
        if parent_cap:
            parent_cap = parent_cap.type
            while parent_cap != self.TOSCA_TYPEURI_CAPABILITY_ROOT:
                if custom_def and parent_cap in custom_def.keys():
                    capabilities[parent_cap] = custom_def[parent_cap]
                elif parent_cap in self.TOSCA_DEF.keys():
                    capabilities[parent_cap] = self.TOSCA_DEF[parent_cap]
                parent_cap = capabilities[parent_cap]['derived_from']
        return capabilities

    @property
    def parent_type(self):
        '''Return a capability this capability is derived from.'''
        if not hasattr(self, 'defs'):
            return None
        pnode = self.derived_from(self.defs)
        if pnode:
            return CapabilityTypeDef(self.name, pnode,
                                     self.nodetype, self.custom_def)

    def inherits_from(self, type_names):
        '''Check this capability is in type_names

           Check if this capability or some of its parent types
           are in the list of types: type_names
        '''
        if self.type in type_names:
            return True
        elif self.parent_type:
            return self.parent_type.inherits_from(type_names)
        else:
            return False



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from toscaparser.tests.base import TestCase
import toscaparser.utils.urlutils
import toscaparser.utils.yamlparser

YAML_LOADER = toscaparser.utils.yamlparser.load_yaml


class UrlUtilsTest(TestCase):

    url_utils = toscaparser.utils.urlutils.UrlUtils

    def test_urlutils_validate_url(self):
        self.assertTrue(self.url_utils.validate_url("http://www.github.com/"))
        self.assertTrue(
            self.url_utils.validate_url("https://github.com:81/a/2/a.b"))
        self.assertTrue(self.url_utils.validate_url("ftp://github.com"))
        self.assertFalse(self.url_utils.validate_url("github.com"))
        self.assertFalse(self.url_utils.validate_url("123"))
        self.assertFalse(self.url_utils.validate_url("a/b/c"))
        self.assertTrue(self.url_utils.validate_url("file:///dir/file.ext"))

    def test_urlutils_join_url(self):
        self.assertEqual(
            self.url_utils.join_url("http://github.com/proj1", "proj2"),
            "http://github.com/proj2")
        self.assertEqual(
            self.url_utils.join_url("http://github.com/proj1/scripts/a.js",
                                    "b.js"),
            "http://github.com/proj1/scripts/b.js")
        self.assertEqual(
            self.url_utils.join_url("http://github.com/proj1/scripts", "b.js"),
            "http://github.com/proj1/b.js")
        self.assertEqual(
            self.url_utils.join_url("http://github.com/proj1/scripts",
                                    "scripts/b.js"),
            "http://github.com/proj1/scripts/b.js")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from toscaparser.common.exception import ExceptionCollector
from toscaparser.common.exception import UnknownFieldError
from toscaparser.elements.statefulentitytype import StatefulEntityType

SECTIONS = (LIFECYCLE, CONFIGURE, LIFECYCLE_SHORTNAME,
            CONFIGURE_SHORTNAME) = \
           ('tosca.interfaces.node.lifecycle.Standard',
            'tosca.interfaces.relationship.Configure',
            'Standard', 'Configure')

INTERFACEVALUE = (IMPLEMENTATION, INPUTS, OUTPUTS) = \
    ('implementation', 'inputs', 'outputs')

INTERFACE_DEF_RESERVED_WORDS = ['type', 'inputs', 'derived_from', 'version',
                                'description']


class InterfacesDef(StatefulEntityType):
    '''TOSCA built-in interfaces type.'''

    def __init__(self, node_type, interfacename,
                 node_template=None, name=None, value=None):
        self.ntype = node_type
        self.node_template = node_template
        self.type = interfacename
        self.interfacename = interfacename
        self.name = name
        self.value = value
        self.implementation = None
        self.inputs = None
        self.outputs = None
        self.defs = {}
        if interfacename == LIFECYCLE_SHORTNAME:
            self.interfacetype = LIFECYCLE
        elif interfacename == CONFIGURE_SHORTNAME:
            self.interfacetype = CONFIGURE
        elif hasattr(self.ntype, 'interfaces') \
                and self.ntype.interfaces \
                and interfacename in self.ntype.interfaces:
            self.interfacetype = self.ntype.interfaces[interfacename]['type']
        if not self.interfacetype:
            ExceptionCollector.appendException(
                TypeError("Interface type for interface \"{0}\" not found"
                          .format(self.interfacename))
            )
        if node_type:
            if self.node_template and self.node_template.custom_def \
               and self.interfacetype in self.node_template.custom_def:
                self.defs = self.node_template.custom_def[self.interfacetype]
            elif self.interfacetype in self.TOSCA_DEF:
                self.defs = self.TOSCA_DEF[self.interfacetype]
        if not self.defs:
            ExceptionCollector.appendException(
                TypeError("Interface type definition for interface \"{0}\" "
                          "not found".format(self.interfacetype))
            )
        if value:
            if isinstance(self.value, dict):
                for i, j in self.value.items():
                    if i == IMPLEMENTATION:
                        self.implementation = j
                    elif i == INPUTS:
                        self.inputs = j
                    elif i == OUTPUTS:
                        self.outputs = j
                    else:
                        what = ('"interfaces" of template "%s"' %
                                self.node_template.name)
                        ExceptionCollector.appendException(
                            UnknownFieldError(what=what, field=i))
            else:
                self.implementation = value

    @property
    def lifecycle_ops(self):
        if self.defs:
            if self.type == LIFECYCLE:
                return self._ops()

    @property
    def configure_ops(self):
        if self.defs:
            if self.type == CONFIGURE:
                return self._ops()

    def _ops(self):
        ops = []
        for name in list(self.defs.keys()):
            ops.append(name)
        return ops



# Copyright (C) 2024 Fujitsu
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os


def get_sample_dir():
    return os.path.abspath(os.path.join(os.path.dirname(__file__),
                                        '../../samples'))


def get_sample_test_dir():
    # {tosca-parser}/samples/tests
    return os.path.join(get_sample_dir(), 'tests')


def get_sample_test_path(*p):
    return os.path.join(get_sample_test_dir(), *p)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from toscaparser.common.exception import ExceptionCollector
from toscaparser.common.exception import InvalidTypeError
from toscaparser.common.exception import UnknownFieldError
from toscaparser.elements.statefulentitytype import StatefulEntityType


class GroupType(StatefulEntityType):
    '''TOSCA built-in group type.'''

    # In TOSCA1.3, the section name 'interfaces' is not defined for group type.
    # It will be ignored if the 'interfaces' is defined as TOSCA1.3.
    SECTIONS = (DERIVED_FROM, VERSION, METADATA, DESCRIPTION, PROPERTIES,
                MEMBERS, INTERFACES, ATTRIBUTES) = \
               ("derived_from", "version", "metadata", "description",
                "properties", "members", "interfaces", "attributes")

    def __init__(self, grouptype, custom_def=None):
        super(GroupType, self).__init__(grouptype, self.GROUP_PREFIX,
                                        custom_def)
        self.custom_def = custom_def
        self.grouptype = grouptype
        self._validate_fields()
        self.group_description = None
        if self.DESCRIPTION in self.defs:
            self.group_description = self.defs[self.DESCRIPTION]

        self.group_version = None
        if self.VERSION in self.defs:
            self.group_version = self.defs[self.VERSION]

        self.group_properties = None
        if self.PROPERTIES in self.defs:
            self.group_properties = self.defs[self.PROPERTIES]

        self.group_members = None
        if self.MEMBERS in self.defs:
            self.group_members = self.defs[self.MEMBERS]

        if self.METADATA in self.defs:
            self.meta_data = self.defs[self.METADATA]
            self._validate_metadata(self.meta_data)

    @property
    def parent_type(self):
        '''Return a group statefulentity of this entity is derived from.'''
        if not hasattr(self, 'defs'):
            return None
        pgroup_entity = self.derived_from(self.defs)
        if pgroup_entity:
            return GroupType(pgroup_entity, self.custom_def)

    @property
    def description(self):
        return self.group_description

    @property
    def version(self):
        return self.group_version

    @property
    def interfaces(self):
        return self.get_value(self.INTERFACES)

    def _validate_fields(self):
        if self.defs:
            for name in self.defs.keys():
                if name not in self.SECTIONS:
                    ExceptionCollector.appendException(
                        UnknownFieldError(what='Group Type %s'
                                          % self.grouptype, field=name))

    def _validate_metadata(self, meta_data):
        if not meta_data.get('type') in ['map', 'tosca:map']:
            ExceptionCollector.appendException(
                InvalidTypeError(what='"%s" defined in group for '
                                 'metadata' % (meta_data.get('type'))))
        for entry_schema, entry_schema_type in meta_data.items():
            if isinstance(entry_schema_type, dict) and not \
                    entry_schema_type.get('type') == 'string':
                ExceptionCollector.appendException(
                    InvalidTypeError(what='"%s" defined in group for '
                                     'metadata "%s"'
                                     % (entry_schema_type.get('type'),
                                        entry_schema)))
